#!/usr/bin/env python3

"""
Archey main file.
It loads each entry as a different class coming from the `entries` module.
Logos are stored under the `logos` module.
"""

import argparse
import os

from enum import Enum
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack

from archey._version import __version__
from archey.output import Output
from archey.configuration import Configuration
from archey.distributions import Distributions
from archey.processes import Processes
from archey.screenshot import take_screenshot
from archey.entries.user import User as e_User
from archey.entries.hostname import Hostname as e_Hostname
from archey.entries.model import Model as e_Model
from archey.entries.distro import Distro as e_Distro
from archey.entries.kernel import Kernel as e_Kernel
from archey.entries.uptime import Uptime as e_Uptime
from archey.entries.processes import Processes as e_Processes
from archey.entries.window_manager import WindowManager as e_WindowManager
from archey.entries.desktop_environment import DesktopEnvironment as e_DesktopEnvironment
from archey.entries.shell import Shell as e_Shell
from archey.entries.terminal import Terminal as e_Terminal
from archey.entries.packages import Packages as e_Packages
from archey.entries.temperature import Temperature as e_Temperature
from archey.entries.cpu import CPU as e_CPU
from archey.entries.gpu import GPU as e_GPU
from archey.entries.ram import RAM as e_RAM
from archey.entries.disk import Disk as e_Disk
from archey.entries.lan_ip import LanIp as e_LanIp
from archey.entries.wan_ip import WanIp as e_WanIp


class Entries(Enum):
    """
    An enumeration to store and declare each one of our entries.
    The string representation of keys will act as entries names.
    Values will be set under the `value` attribute of each obtained objects.
    """
    User = e_User
    Hostname = e_Hostname
    Model = e_Model
    Distro = e_Distro
    Kernel = e_Kernel
    Uptime = e_Uptime
    Processes = e_Processes
    WindowManager = e_WindowManager
    DesktopEnvironment = e_DesktopEnvironment
    Shell = e_Shell
    Terminal = e_Terminal
    Packages = e_Packages
    Temperature = e_Temperature
    CPU = e_CPU
    GPU = e_GPU
    RAM = e_RAM
    Disk = e_Disk
    LAN_IP = e_LanIp
    WAN_IP = e_WanIp


def args_parsing():
    """Simple wrapper to `argparse`"""
    parser = argparse.ArgumentParser(prog='archey')
    parser.add_argument(
        '-c', '--config-path',
        metavar='PATH',
        help='path to a configuration file, or a directory containing a `config.json`'
    )
    parser.add_argument(
        '-d', '--distribution',
        metavar='IDENTIFIER',
        choices=Distributions.get_distribution_identifiers(),
        help='supported distribution identifier to show the logo of, pass `unknown` to list them'
    )
    parser.add_argument(
        '-j', '--json',
        action='count',
        help='output entries data to JSON format, use multiple times to increase indentation'
    )
    parser.add_argument(
        '-s', '--screenshot',
        metavar='PATH',
        nargs='?',
        const=False,
        help='take a screenshot once execution is done, optionally specify a target path'
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=__version__
    )
    return parser.parse_args()


def main():
    """Simple entry point"""
    args = args_parsing()

    # `Processes` is a singleton, let's populate the internal list here.
    Processes()

    # `Configuration` is a singleton, let's populate the internal object here.
    configuration = Configuration(config_path=args.config_path)

    # From configuration, gather the entries user-enabled.
    enabled_entries = [
        (entry.value, entry.name)
        for entry in Entries
        if configuration.get('entries', {}).get(entry.name, True)
    ]

    output = Output(
        preferred_distribution=args.distribution,
        format_to_json=args.json
    )

    # We will map this function onto our enabled entries to instantiate them.
    def _entry_instantiator(entry_tuple):
        return entry_tuple[0](name=entry_tuple[1])

    # Let's use a context manager stack to manage conditional use of `TheadPoolExecutor`.
    with ExitStack() as cm_stack:
        if not configuration.get('parallel_loading'):
            mapper = map
        else:
            # Instantiate a threads pool to load our enabled entries in parallel.
            # We use threads (and not processes) since most work done by our entries is IO-bound.
            # `max_workers` is manually computed to mimic Python 3.8+ behaviour, but for our needs.
            #   See <https://github.com/python/cpython/pull/13618>.
            executor = cm_stack.enter_context(ThreadPoolExecutor(
                max_workers=min(len(enabled_entries) or 1, (os.cpu_count() or 1) + 4)
            ))
            mapper = executor.map

        for entry_instance in mapper(_entry_instantiator, enabled_entries):
            output.add_entry(entry_instance)

    output.output()

    # Has the screenshot flag been specified ?
    if args.screenshot is not None:
        # If so, but still _falsy_, pass `None` as no output file has been specified by the user.
        take_screenshot((args.screenshot or None))


if __name__ == '__main__':
    main()



"""Entry base class"""

from abc import ABC as AbstractBaseClass, abstractmethod

from archey.configuration import Configuration


class Entry(AbstractBaseClass):
    """Module base class"""
    @abstractmethod
    def __init__(self, name=None, value=None):
        # Each entry will have always have the following attributes...
        # `name` (key);
        # `value` (value of entry as an appropriate object)
        # ... which are `None` by default.
        self.name = name
        self.value = value

        # Propagates a reference to `Configuration` singleton to each inheriting class.
        self._configuration = Configuration()


    def output(self, output):
        """Output the results to output. Can be overridden by subclasses."""
        if self.value:
            # Let's assume we can just use `__str__` on the object in value,
            # and create a single-line output with it.
            output.append(self.name, str(self.value))
        else:
            # If the value is "falsy" leave a generic "Not detected" message for this entry.
            output.append(
                self.name,
                self._configuration.get('default_strings')['not_detected']
            )



"""Local IP addresses detection class"""

import sys

try:
    import netifaces
except ImportError:
    netifaces = None

from archey.entry import Entry


class LanIp(Entry):
    """Relies on the `netifaces` module to detect LAN IP addresses"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not netifaces:
            print(
                """\
Warning: `netifaces` Python module couldn\'t be found.
Please either install it or disable `LAN_IP` entry in configuration.\
""",
                file=sys.stderr
            )
            return

        address_types = [netifaces.AF_INET]
        if self._configuration.get('ip_settings')['lan_ip_v6_support']:
            address_types.append(netifaces.AF_INET6)

        self.value = []

        # Loop through all available network interfaces.
        for if_name in netifaces.interfaces():
            # Fetch associated addresses elements.
            if_addrs = netifaces.ifaddresses(if_name)

            # For each IPv4 (or IPv6 address)...
            for addr_type in address_types:
                if addr_type not in if_addrs:
                    continue

                for if_addr in if_addrs[addr_type]:
                    # Filter out loopback addresses.
                    if (addr_type == netifaces.AF_INET and if_addr['addr'].startswith('127.')) \
                        or if_addr['addr'] == '::1':
                        continue

                    self.value.append(if_addr['addr'].split('%')[0])

        lan_ip_max_count = self._configuration.get('ip_settings')['lan_ip_max_count']
        if lan_ip_max_count is not False:
            self.value = self.value[:lan_ip_max_count]


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting the IP address list."""
        # If we found IP addresses, join them together nicely.
        # If not, fall back on default strings according to `netifaces` availability.
        if self.value:
            text_output = ', '.join(self.value)
        elif netifaces:
            text_output = self._configuration.get('default_strings')['no_address']
        else:
            text_output = self._configuration.get('default_strings')['not_detected']

        output.append(self.name, text_output)



"""Test module for Archey's CPU detection module"""

import unittest
from unittest.mock import mock_open, patch

from archey.entries.cpu import CPU


class TestCPUEntry(unittest.TestCase):
    """
    Here, we mock the `open` call on `/proc/cpuinfo` with fake content.
    """
    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
"""),
        create=True
    )
    def test_model_name_match_cpuinfo(self):
        """Test `/proc/cpuinfo` parsing"""
        self.assertEqual(CPU().value, 'CPU-MODEL-NAME')

    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
"""),
        create=True
    )
    @patch(
        'archey.entries.cpu.check_output',
        return_value="""\
Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              4
On-line CPU(s) list: 0-3
Thread(s) per core:  X
Core(s) per socket:  Y
Socket(s):           1
NUMA node(s):        1
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          CPU-MODEL-NAME-WITHOUT-PROC-CPUINFO
""")
    def test_model_name_match_lscpu(self, _):
        """
        Test model name parsing from `lscpu` output.

        See issue #29 (ARM architectures).
        `/proc/cpuinfo` will not contain `model name` info.
        `lscpu` output will be used instead.
        """
        self.assertEqual(CPU().value, 'CPU-MODEL-NAME-WITHOUT-PROC-CPUINFO')

    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU  MODEL\t  NAME
"""),
        create=True
    )
    def test_spaces_squeezing(self):
        """Test name sanitizing, needed on some platforms"""
        self.assertEqual(CPU().value, 'CPU MODEL NAME')

    @patch(
        'archey.entries.cpu.open',
        side_effect=PermissionError(),
        create=True
    )
    @patch(
        'archey.entries.cpu.check_output',
        return_value="""\
Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              4
On-line CPU(s) list: 0-3
Thread(s) per core:  X
Core(s) per socket:  Y
Socket(s):           1
NUMA node(s):        1
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          CPU-MODEL-NAME-WITHOUT-PROC-CPUINFO
""")
    def test_proc_cpuinfo_unreadable(self, _, __):
        """Check Archey does not crash when `/proc/cpuinfo` is not readable"""
        self.assertEqual(CPU().value, 'CPU-MODEL-NAME-WITHOUT-PROC-CPUINFO')


if __name__ == '__main__':
    unittest.main()



"""Windows manager detection class"""

import re

from subprocess import check_output, DEVNULL, CalledProcessError

from archey.entry import Entry
from archey.processes import Processes


WM_DICT = {
    'awesome': 'Awesome',
    'beryl': 'Beryl',
    'blackbox': 'Blackbox',
    'bspwm': 'bspwm',
    'cinnamon': 'Cinnamon',
    'compiz': 'Compiz',
    'deepin-wm': 'Deepin WM',
    'dwm': 'DWM',
    'enlightenment': 'Enlightenment',
    'herbstluftwm': 'herbstluftwm',
    'fluxbox': 'Fluxbox',
    'fvwm': 'FVWM',
    'i3': 'i3',
    'icewm': 'IceWM',
    'kwin_x11': 'KWin',
    'metacity': 'Metacity',
    'musca': 'Musca',
    'openbox': 'Openbox',
    'pekwm': 'PekWM',
    'qtile': 'QTile',
    'ratpoison': 'ratpoison',
    'scrotwm': 'ScrotWM',
    'stumpwm': 'StumpWM',
    'subtle': 'Subtle',
    'monsterwm': 'MonsterWM',
    'wingo': 'Wingo',
    'wmaker': 'Window Maker',
    'wmfs': 'Wmfs',
    'wmii': 'wmii',
    'xfwm4': 'Xfwm',
    'xmonad': 'xmonad'
}


class WindowManager(Entry):
    """
    Uses `wmctrl` to retrieve some information about the window manager.
    If not available, fall back on a simple iteration over the processes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.value = re.search(
                '(?<=Name: ).*',
                check_output(
                    ['wmctrl', '-m'],
                    stderr=DEVNULL, universal_newlines=True
                )
            ).group(0)
        except (FileNotFoundError, CalledProcessError):
            processes = Processes().list
            for wm_id, wm_name in WM_DICT.items():
                if wm_id in processes:
                    self.value = wm_name
                    break



"""Test module for Archey's public IP address detection module"""

import unittest
from unittest.mock import MagicMock, patch

from socket import timeout as SocketTimeoutError
from subprocess import TimeoutExpired
from urllib.error import URLError

from archey.test import CustomAssertions
from archey.entries.wan_ip import WanIp
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG

class TestWanIpEntry(unittest.TestCase, CustomAssertions):
    """
    Here, we mock calls to `dig` or `urlopen`.
    """
    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=[
            'XXX.YY.ZZ.TTT\n',
            '0123::4567:89a:dead:beef\n'
        ]
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': True
            }
        }
    )
    def test_ipv6_and_ipv4(self, _):
        """Test the regular case : Both IPv4 and IPv6 are retrieved"""
        self.assertEqual(
            WanIp().value,
            ['XXX.YY.ZZ.TTT', '0123::4567:89a:dead:beef']
        )

    @patch(
        'archey.entries.wan_ip.check_output',
        return_value='XXX.YY.ZZ.TTT'
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': False
            }
        }
    )
    def test_ipv4_only(self, _):
        """Test only public IPv4 detection"""
        self.assertEqual(
            WanIp().value,
            ['XXX.YY.ZZ.TTT']
        )

    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=[
            'XXX.YY.ZZ.TTT',            # The IPv4 address is detected
            TimeoutExpired('dig', 1)  # `check_output` call will fail
        ]
    )
    @patch('archey.entries.wan_ip.urlopen')
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': True
            }
        }
    )
    def test_ipv6_timeout(self, urlopen_mock, _):
        """
        Test when `dig` call timeout for the IPv6 detection.
        Additionally check the `output` method behavior.
        """
        urlopen_mock.return_value.read.return_value = b'0123::4567:89a:dead:beef\n'

        wan_ip = WanIp()

        output_mock = MagicMock()
        wan_ip.output(output_mock)

        self.assertListEqual(
            wan_ip.value,
            ['XXX.YY.ZZ.TTT', '0123::4567:89a:dead:beef']
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            'XXX.YY.ZZ.TTT, 0123::4567:89a:dead:beef'
        )

    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=TimeoutExpired('dig', 1)  # `check_output` call will fail
    )
    @patch(
        'archey.entries.wan_ip.urlopen',
        side_effect=URLError('<urlopen error timed out>')  # `urlopen` call will fail
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': False
            }
        }
    )
    def test_ipv4_timeout_twice(self, _, __):
        """Test when both `dig` and `URLOpen` trigger timeouts..."""
        self.assertListEmpty(WanIp().value)

    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=TimeoutExpired('dig', 1)  # `check_output` call will fail
    )
    @patch(
        'archey.entries.wan_ip.urlopen',
        side_effect=SocketTimeoutError(1)  # `urlopen` call will fail
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': False
            }
        }
    )
    def test_ipv4_timeout_twice_socket_error(self, _, __):
        """Test when both `dig` timeouts and `URLOpen` raises `socket.timeout`..."""
        self.assertListEmpty(WanIp().value)

    @patch(
        'archey.entries.wan_ip.check_output',
        return_value=''  # No address will be returned
    )
    @patch(
        'urllib.request.urlopen',
        return_value=None  # No object will be returned
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'wan_ip_v6_support': False
            }
        }
    )
    def test_no_address(self, _, __):
        """
        Test when no address could be retrieved.
        Additionally check the `output` method behavior.
        """
        wan_ip = WanIp()

        output_mock = MagicMock()
        wan_ip.output(output_mock)

        self.assertListEmpty(wan_ip.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['no_address']
        )


if __name__ == '__main__':
    unittest.main()



"""Kernel information detection class"""

from subprocess import check_output

from archey.entry import Entry


class Kernel(Entry):
    """Another call to `uname` to retrieve kernel release information"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = check_output(
            ['uname', '-r'],
            universal_newlines=True
        ).rstrip()



"""Simple singleton meta-class definition"""


from abc import ABCMeta as AbstractBaseMetaClass


class Singleton(AbstractBaseMetaClass):
    """
    Taken from : <https://stackoverflow.com/q/6760685/10599709>
    This meta-class allows us to declare `Configuration` as a singleton.
    This way, we are able to import `Configuration` in multiple modules, ...
    ... whereas it is effectively loaded only once.
    You cannot instantiate this meta-class directly.
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""Test module for Archey's LAN IP addresses detection module"""

import unittest
from unittest.mock import MagicMock, patch

from netifaces import AF_INET, AF_INET6, AF_LINK

from archey.entries.lan_ip import LanIp
from archey.test import CustomAssertions
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestLanIpEntry(unittest.TestCase, CustomAssertions):
    """Here, we mock the `netifaces` usages (interfaces and addresses detection calls)"""
    @patch(
        'archey.entries.lan_ip.netifaces.interfaces',
        return_value=['lo', 'en0', 'wlo1']
    )
    @patch(
        'archey.entries.lan_ip.netifaces.ifaddresses',
        side_effect=[
            {
                AF_INET: [{
                    'addr': '127.0.0.1',
                    'netmask': '255.0.0.0'
                }]
            },
            {
                AF_INET: [
                    {
                        'addr': '192.168.0.11',
                        'netmask': '255.255.255.0',
                        'broadcast': '192.168.0.255'
                    },
                    {
                        'addr': '192.168.1.11',
                        'netmask': '255.255.255.0',
                        'broadcast': '192.168.1.255'
                    }
                ]
            },
            {
                AF_INET: [{
                    'addr': '172.34.56.78',
                    'broadcast': '172.34.255.255'
                }]
            }
        ]
    )
    @HelperMethods.patch_clean_configuration(
        configuration={'ip_settings': {'lan_ip_max_count': False}}
    )
    def test_multiple_interfaces(self, _, __):
        """Test for multiple interfaces, multiple addresses (including a loopback one)"""
        self.assertListEqual(
            LanIp().value,
            ['192.168.0.11', '192.168.1.11', '172.34.56.78']
        )

    @patch(
        'archey.entries.lan_ip.netifaces.interfaces',
        return_value=['lo', 'en0']
    )
    @patch(
        'archey.entries.lan_ip.netifaces.ifaddresses',
        side_effect=[
            {
                AF_LINK: [{
                    'addr': '00:00:00:00:00:00',
                    'peer': '00:00:00:00:00:00'
                }],
                AF_INET: [{
                    'addr': '127.0.0.1',
                    'netmask': '255.0.0.0',
                    'peer': '127.0.0.1'
                }],
                AF_INET6: [{
                    'addr': '::1',
                    'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff/128'
                }]
            },
            {
                AF_LINK: [{
                    'addr': 'de:ad:be:ef:de:ad',
                    'broadcast': 'ff:ff:ff:ff:ff:ff'
                }],
                AF_INET: [{
                    'addr': '192.168.1.55',
                    'netmask': '255.255.255.0',
                    'broadcast': '192.168.1.255'
                }],
                AF_INET6: [
                    {
                        'addr': '2001::45:6789:abcd:6817',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    },
                    {
                        'addr': '2a02::45:6789:abcd:0123/64',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    },
                    {
                        'addr': r'fe80::abcd:ef0:abef:dead\%en0',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    }
                ]
            }
        ]
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'ip_settings': {
                'lan_ip_max_count': 2,
                'lan_ip_v6_support': True
            }
        }
    )
    def test_ipv6_and_limit_and_ether(self, _, __):
        """
        Test for IPv6 support, final set length limit and Ethernet interface filtering.
        Additionally check the `output` method behavior.
        """
        lan_ip = LanIp()

        output_mock = MagicMock()
        lan_ip.output(output_mock)

        self.assertListEqual(
            lan_ip.value,
            ['192.168.1.55', '2001::45:6789:abcd:6817']
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '192.168.1.55, 2001::45:6789:abcd:6817'
        )

    @patch(
        'archey.entries.lan_ip.netifaces.interfaces',
        return_value=['lo', 'en0']
    )
    @patch(
        'archey.entries.lan_ip.netifaces.ifaddresses',
        side_effect=[
            {
                AF_INET: [{
                    'addr': '127.0.0.1',
                    'netmask': '255.0.0.0',
                    'peer': '127.0.0.1'
                }],
                AF_INET6: [{
                    'addr': '::1',
                    'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff/128'
                }]
            },
            {
                AF_INET: [{
                    'addr': '192.168.1.55',
                    'netmask': '255.255.255.0',
                    'broadcast': '192.168.1.255'
                }],
                AF_INET6: [
                    {
                        'addr': '2001::45:6789:abcd:6817',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    },
                    {
                        'addr': '2a02::45:6789:abcd:0123/64',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    },
                    {
                        'addr': r'fe80::abcd:ef0:abef:dead\%en0',
                        'netmask': 'ffff:ffff:ffff:ffff::/64'
                    }
                ]
            }
        ]
    )
    @HelperMethods.patch_clean_configuration(
        configuration={'ip_settings': {'lan_ip_v6_support': False}}
    )
    def test_no_ipv6(self, _, __):
        """Test for IPv6 hiding"""
        self.assertListEqual(
            LanIp().value,
            ['192.168.1.55']
        )

    @patch(
        'archey.entries.lan_ip.netifaces.interfaces',
        return_value=[]  # No interface returned by `netifaces`.
    )
    @HelperMethods.patch_clean_configuration
    def test_no_network_interface(self, _):
        """Test when the device does not have any network interface"""
        self.assertListEmpty(LanIp().value)

    @patch(
        'archey.entries.lan_ip.netifaces.interfaces',
        return_value=['lo', 'en0']
    )
    @patch(
        'archey.entries.lan_ip.netifaces.ifaddresses',
        side_effect=[
            {
                AF_LINK: [{
                    'addr': '00:00:00:00:00:00',
                    'peer': '00:00:00:00:00:00'
                }],
                AF_INET: [{
                    'addr': '127.0.0.1',
                    'netmask': '255.0.0.0',
                    'peer': '127.0.0.1'
                }],
                AF_INET6: [{
                    'addr': '::1',
                    'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff/128'
                }]
            },
            {
                # No address for this one.
            }
        ]
    )
    @HelperMethods.patch_clean_configuration
    def test_no_network_address_output(self, _, __):
        """
        Test when the network interface(s) do not have any IP address.
        Additionally check the `output` method behavior.
        """
        lan_ip = LanIp()

        output_mock = MagicMock()
        lan_ip.output(output_mock)

        self.assertListEmpty(lan_ip.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['no_address']
        )

    @patch(
        'archey.entries.lan_ip.netifaces',
        None  # Imitate an `ImportError` behavior.
    )
    @patch(
        'archey.entries.lan_ip.print',
        return_value=None,  # Let's nastily mute class' outputs.
        create=True
    )
    @HelperMethods.patch_clean_configuration
    def test_netifaces_not_available(self, _):
        """Check `netifaces` is really acting as a (soft-)dependency"""
        lan_ip = LanIp()

        output_mock = MagicMock()
        lan_ip.output(output_mock)

        self.assertIsNone(lan_ip.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""GPU information detection class"""

from subprocess import CalledProcessError, check_output
from itertools import islice

from archey.entry import Entry


class GPU(Entry):
    """Relies on `lspci` to retrieve graphical device(s) information"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        max_count = self._configuration.get('gpu')['max_count']
        # Consistency with other entries' configuration: Infinite count if false.
        if max_count is False:
            max_count = None

        # Populate our list of devices with the `lspci`-based generator.
        self.value = list(islice(self._gpu_generator(), max_count))

    @staticmethod
    def _gpu_generator():
        """Based on `lspci` output, return a generator for video controllers names"""
        try:
            lspci_output = check_output(
                ['lspci'],
                universal_newlines=True
            ).splitlines()
        except (FileNotFoundError, CalledProcessError):
            return

        # We'll be looking for specific video controllers (in the below keys order).
        for video_key in ('3D', 'VGA', 'Display'):
            for pci_device in lspci_output:
                # If a controller type match...
                if video_key in pci_device:
                    # ... return its name on the next iteration.
                    yield pci_device.partition(': ')[2]


    def output(self, output):
        """Writes GPUs to `output` based on preferences"""
        # Even when no GPU device could be detected, be sure to add an "empty" entry to `output`.
        if self._configuration.get('gpu')['one_line'] or not self.value:
            output.append(
                self.name,
                ', '.join(self.value) or self._configuration.get('default_strings')['not_detected']
            )
        else:
            for gpu_device in self.value:
                output.append(self.name, gpu_device)



"""Test module for Archey's window manager detection module"""

import unittest
from unittest.mock import MagicMock, patch

from archey.entries.window_manager import WindowManager
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestWindowManagerEntry(unittest.TestCase):
    """
    Here, we mock the `check_output` call and check afterwards
      that the output is correct.
    We've to test the case where `wmctrl` is not installed too.
    """
    @patch(
        'archey.entries.window_manager.check_output',
        return_value="""\
Name: WINDOW MANAGER
Class: N/A
PID: N/A
Window manager's "showing the desktop" mode: OFF
""")
    def test_wmctrl(self, _):
        """Test `wmctrl` output parsing"""
        self.assertEqual(WindowManager().value, 'WINDOW MANAGER')

    @patch(
        'archey.entries.window_manager.check_output',
        side_effect=FileNotFoundError()  # `wmctrl` call will fail
    )
    @patch(
        'archey.entries.window_manager.Processes.list',
        (  # Fake running processes list
            'some',
            'awesome',  # Match !
            'programs',
            'running',
            'here'
        )
    )
    def test_no_wmctrl_match(self, _):
        """Test basic detection based on a (fake) processes list"""
        self.assertEqual(WindowManager().value, 'Awesome')

    @patch(
        'archey.entries.window_manager.check_output',
        side_effect=FileNotFoundError()  # `wmctrl` call will fail
    )
    @patch(
        'archey.entries.window_manager.Processes.list',
        (  # Fake running processes list
            'some',
            'weird',  # Mismatch !
            'programs',
            'running',
            'here'
        )
    )
    @HelperMethods.patch_clean_configuration
    def test_no_wmctrl_mismatch(self, _):
        """Test (non-detection) when processes list do not contain any known value"""
        window_manager = WindowManager()

        output_mock = MagicMock()
        window_manager.output(output_mock)

        self.assertIsNone(window_manager.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""Alpine Linux logo"""

ALPINE_LINUX = [
    """{c[0]}        .hddddddddddddddddddddddh.          """,
    """{c[0]}       :dddddddddddddddddddddddddd:         """,
    """{c[0]}      /dddddddddddddddddddddddddddd/        """,
    """{c[0]}     +dddddddddddddddddddddddddddddd+       """,
    """{c[0]}   `sdddddddddddddddddddddddddddddddds`     """,
    """{c[0]}  `ydddddddddddd++hdddddddddddddddddddy`    """,
    """{c[0]} .hddddddddddd+`  `+ddddh:-sdddddddddddh.   """,
    """{c[0]} hdddddddddd+`      `+y:    .sddddddddddh   """,
    """{c[0]} ddddddddh+`   `//`   `.`     -sddddddddd   """,
    """{c[0]} ddddddh+`   `/hddh/`   `:s-    -sddddddd   """,
    """{c[0]} ddddh+`   `/+/dddddh/`   `+s-    -sddddd   """,
    """{c[0]} ddd+`   `/o` :dddddddh/`   `oy-    .yddd   """,
    """{c[0]} hdddyo+ohddyosdddddddddho+oydddy++ohdddh   """,
    """{c[0]} .hddddddddddddddddddddddddddddddddddddh.   """,
    """{c[0]}  `yddddddddddddddddddddddddddddddddddy`    """,
    """{c[0]}   `sdddddddddddddddddddddddddddddddds`     """,
    """{c[0]}     +dddddddddddddddddddddddddddddd+       """,
    """{c[0]}      /dddddddddddddddddddddddddddd/        """,
    """{c[0]}       :dddddddddddddddddddddddddd:         """,
    """{c[0]}        .hddddddddddddddddddddddh.          """
]



"""
Logos colors definition.
Distributions / logos correspondences.
Default configuration.
"""

import archey.logos as logos

from archey.colors import Colors
from archey.distributions import Distributions


# This dictionary contains which colors should be used for each supported distribution.
# The first element (`[0]`) of each list will be used to display text entries.
COLORS_DICT = {
    Distributions.ARCH_LINUX: [Colors.CYAN_BRIGHT, Colors.CYAN_NORMAL],
    Distributions.ANDROID: [Colors.GREEN_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.ALPINE_LINUX: [Colors.BLUE_BRIGHT],
    Distributions.BUNSENLABS: [Colors.WHITE_BRIGHT, Colors.YELLOW_BRIGHT, Colors.YELLOW_NORMAL],
    Distributions.CENTOS: [
        Colors.WHITE_BRIGHT,
        Colors.YELLOW_NORMAL, Colors.GREEN_BRIGHT, Colors.BLUE_NORMAL, Colors.MAGENTA_BRIGHT
    ],
    Distributions.CRUNCHBANG: [Colors.WHITE_BRIGHT],
    Distributions.DEBIAN: [Colors.RED_BRIGHT, Colors.RED_NORMAL],
    Distributions.ELEMENTARY: [Colors.WHITE_BRIGHT],
    Distributions.FEDORA: [Colors.BLUE_BRIGHT, Colors.BLUE_NORMAL, Colors.WHITE_BRIGHT],
    Distributions.FREEBSD: [Colors.RED_BRIGHT, Colors.RED_NORMAL],
    Distributions.GENTOO: [Colors.MAGENTA_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.KALI_LINUX: [Colors.BLUE_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.MANJARO_LINUX: [Colors.GREEN_BRIGHT],
    Distributions.NIXOS: [Colors.BLUE_NORMAL, Colors.CYAN_NORMAL],
    Distributions.LINUX: [Colors.WHITE_BRIGHT, Colors.YELLOW_BRIGHT],
    Distributions.LINUX_MINT: [Colors.GREEN_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.POP: [Colors.CYAN_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.OPENBSD: [Colors.YELLOW_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.OPENSUSE: [Colors.GREEN_NORMAL, Colors.WHITE_BRIGHT],
    Distributions.RASPBIAN: [Colors.RED_BRIGHT, Colors.GREEN_NORMAL],
    Distributions.RED_HAT: [Colors.RED_BRIGHT, Colors.WHITE_BRIGHT, Colors.RED_NORMAL],
    Distributions.SLACKWARE: [Colors.BLUE_NORMAL, Colors.BLUE_BRIGHT, Colors.CLEAR],
    Distributions.UBUNTU: [Colors.RED_BRIGHT, Colors.WHITE_BRIGHT],
    Distributions.WINDOWS: [
        Colors.BLUE_BRIGHT, Colors.RED_BRIGHT, Colors.GREEN_BRIGHT, Colors.YELLOW_NORMAL
    ]
}


# This dictionary contains which logo should be used for each supported distribution.
LOGOS_DICT = {
    Distributions.ALPINE_LINUX: logos.ALPINE_LINUX,
    Distributions.ANDROID: logos.ANDROID,
    Distributions.ARCH_LINUX: logos.ARCH_LINUX,
    Distributions.BUNSENLABS: logos.BUNSENLABS,
    Distributions.CENTOS: logos.CENTOS,
    Distributions.CRUNCHBANG: logos.CRUNCHBANG,
    Distributions.DEBIAN: logos.DEBIAN,
    Distributions.ELEMENTARY: logos.ELEMENTARY,
    Distributions.FEDORA: logos.FEDORA,
    Distributions.FREEBSD: logos.FREEBSD,
    Distributions.GENTOO: logos.GENTOO,
    Distributions.KALI_LINUX: logos.KALI_LINUX,
    Distributions.MANJARO_LINUX: logos.MANJARO,
    Distributions.NIXOS: logos.NIXOS,
    Distributions.LINUX: logos.LINUX,
    Distributions.LINUX_MINT: logos.LINUX_MINT,
    Distributions.POP: logos.POP,
    Distributions.OPENBSD: logos.OPENBSD,
    Distributions.OPENSUSE: logos.OPENSUSE,
    Distributions.RASPBIAN: logos.RASPBIAN,
    Distributions.RED_HAT: logos.RED_HAT,
    Distributions.SLACKWARE: logos.SLACKWARE,
    Distributions.UBUNTU: logos.UBUNTU,
    Distributions.WINDOWS: logos.WINDOWS
}


# The default needed configuration which will be used by Archey is present below.
DEFAULT_CONFIG = {
    'allow_overriding': True,
    'parallel_loading': True,
    'suppress_warnings': False,
    'colors_palette': {
        'use_unicode': True,
        'honor_ansi_color': True
    },
    'disk': {
        'show_filesystems': ['local'],
        'combine_total': True,
        'disk_labels': None,
        'hide_entry_name': None
    },
    'default_strings': {
        'no_address': 'No Address',
        'not_detected': 'Not detected',
        'virtual_environment': 'Virtual Environment'
    },
    'gpu': {
        'one_line': True,
        'max_count': 2
    },
    'ip_settings': {
        'lan_ip_max_count': 2,
        'lan_ip_v6_support': True,
        'wan_ip_v6_support': True
    },
    'limits': {
        'ram': {
            'warning': 33.3,
            'danger': 66.7
        },
        'disk': {
            'warning': 50,
            'danger': 75
        }
    },
    'temperature': {
        'char_before_unit': ' ',
        'sensors_chipsets': [],
        'use_fahrenheit': False
    },
    'timeout': {
        'ipv4_detection': 1,
        'ipv6_detection': 1
    }
}



"""Disk usage detection class"""

import re
from subprocess import CalledProcessError, check_output
from csv import reader as csv_reader

from archey.colors import Colors
from archey.entry import Entry


class Disk(Entry):
    """Uses `df` to compute disk usage across devices"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Populate an output from `df`
        self._disk_dict = self.get_df_output_dict()

        config_filesystems = self._configuration.get('disk')['show_filesystems']
        # See `Disk.get_df_output_dict` for the format we use in `self.value`.
        if config_filesystems == ['local']:
            self.value = self._get_local_filesystems()
        else:
            self.value = self._get_specified_filesystems(config_filesystems)


    def _get_local_filesystems(self):
        """
        Extracts local (i.e. /dev/xxx) filesystems for any *NIX from `self._disk_dict`,
        returning a copy with those filesystems only.

        We specifically ignore...
        Loop devices:-
            /dev(/...)/loop filesystems (Linux)
            /dev(/...)/*vnd filesystems (BSD)
            /dev(/...)/lofi filesystems (Solaris)
        Device mappers:- (since their partitions are already included!)
            /dev(/...)/dm filesystems (Linux)
        """
        # Compile a REGEXP pattern to match the device paths we will accept.
        device_path_regexp = re.compile(r'^\/dev\/(?:(?!loop|[rs]?vnd|lofi|dm).)+$')

        # Build the dictionary
        local_disk_dict = {}
        for mount_point, disk_data in self._disk_dict.items():
            if (
                    device_path_regexp.match(disk_data['device_path'])
                    # De-duplication based on `device_path`s:
                    and not any(
                        disk_data['device_path'] == present_disk_data['device_path']
                        for present_disk_data in local_disk_dict.values()
                    )
            ):
                local_disk_dict[mount_point] = disk_data

        return local_disk_dict


    def _get_specified_filesystems(self, specified_filesystems):
        """
        Extracts the specified filesystems (if found) from `self._disk_dict`,
        returning a copy with those filesystems only, preserving specified mount point names.

        Searching for non-existent filesystems or filesystems based on device paths is slower,
        so try to only search for existing *mount points* if possible!
        """
        specified_disk_dict = {}

        for filesystem in specified_filesystems:
            # Let's use EAFP and first assume the filesystem is a mount point,
            # since the dictionary lookup is O(1).
            try:
                specified_disk_dict[filesystem] = self._disk_dict[filesystem]
                # If we reach here, we found the filesystem - we can move on to the next.
                continue
            except KeyError:
                # It's not a mount point!
                pass

            # Now assume this is a device path.
            for mount_point, disk_data in self._disk_dict.items():
                if disk_data['device_path'] == filesystem:
                    specified_disk_dict[mount_point] = disk_data
                    # We only need one match, so we can stop when it's found.
                    break

        return specified_disk_dict


    @staticmethod
    def get_df_output_dict():
        """
        Runs `df -P -k` and returns disks in a dict formatted as:
        {
            'mount_point_1': {
                'device_path': AAA,
                'used_blocks': BBB,
                'total_blocks': CCC
            },
            'mount_point_2': {
                'device_path': XXX,
                'used_blocks': YYY,
                'total_blocks': ZZZ
            }
        }
        Mount points are used as keys since they are always unique.
        """
        try:
            df_output = check_output(
                ['df', '-P', '-k'],
                env={'LANG': 'C'}, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            # `df` isn't available on this system.
            return {}

        # Parse this output as a table in SSV (space-separated values) format
        ssv_reader = csv_reader(
            df_output.splitlines()[1:],  # Discard the header row here.
            delimiter=' ',
            skipinitialspace=True
        )

        return {
            line[5]: {  # 6th column === mount point.
                'device_path': line[0],
                'used_blocks': int(line[2]),
                'total_blocks': int(line[1])
            } for line in ssv_reader
        }


    @staticmethod
    def _blocks_to_human_readable(blocks, suffix='B'):
        """
        Returns human-readable format of `blocks` supplied in kibibytes (1024 bytes).
        Taken (and modified) from: <https://stackoverflow.com/a/1094933/13343912>
        """
        for unit in ('Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'):
            if blocks < 1024.0:
                break

            blocks /= 1024.0

        return '{0:02.1f} {1}{2}'.format(blocks, unit, suffix)


    def output(self, output):
        """
        Adds the entry to `output` after formatting with color and units.
        Follows the user configuration supplied for formatting.
        """
        # Fetch our `filesystems` object locally so we can modify it safely.
        filesystems = self.value

        if not filesystems:
            # We didn't find any disk, fall back to the default entry behavior.
            super().output(output)
            return

        # DRY configuration object for the output.
        disk_labels = self._configuration.get('disk')['disk_labels']
        hide_entry_name = self._configuration.get('disk')['hide_entry_name']

        # Combine all entries into one grand-total if configured to do so.
        if self._configuration.get('disk')['combine_total']:
            name = self.name

            # Rewrite our `filesystems` object as one combining all of them.
            filesystems = {
                None: {
                    'device_path': None,
                    'used_blocks': sum([
                        filesystem_data['used_blocks']
                        for filesystem_data in filesystems.values()
                    ]),
                    'total_blocks': sum([
                        filesystem_data['total_blocks']
                        for filesystem_data in filesystems.values()
                    ])
                }
            }
        else:
            # We will only use disk labels and entry name hiding if we aren't combining entries.
            name = ''
            # Hide `Disk` from entry name only if the user specified it... as long as a label.
            if not hide_entry_name or not disk_labels:
                name += self.name
            if disk_labels:
                if not hide_entry_name:
                    name += ' '
                name += '({disk_label})'

        # Fetch the user-defined limits from the configuration.
        disk_limits = self._configuration.get('limits')['disk']

        # We will only run this loop a single time for combined entries.
        for mount_point, filesystem_data in filesystems.items():
            # Select the corresponding level color based on disk percentage usage.
            level_color = Colors.get_level_color(
                (filesystem_data['used_blocks'] / filesystem_data['total_blocks']) * 100,
                disk_limits['warning'], disk_limits['danger']
            )

            # Set the correct disk label
            if disk_labels == 'mount_points':
                disk_label = mount_point
            elif disk_labels == 'device_paths':
                disk_label = filesystem_data['device_path']
            else:
                disk_label = None

            pretty_filesystem_value = '{0}{1}{2} / {3}'.format(
                level_color,
                self._blocks_to_human_readable(filesystem_data['used_blocks']),
                Colors.CLEAR,
                self._blocks_to_human_readable(filesystem_data['total_blocks'])
            )

            output.append(
                name.format(disk_label=disk_label),
                pretty_filesystem_value
            )



"""Test module for Archey's RAM usage detection module"""

import unittest
from unittest.mock import mock_open, patch, MagicMock

from archey.colors import Colors
from archey.entries.ram import RAM
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestRAMEntry(unittest.TestCase):
    """
    Here, we mock the `check_output` call to `free` using all three levels of available ram.
    In the last test, mock with `/proc/meminfo` file opening during the manual way.
    """
    @patch(
        'archey.entries.ram.check_output',
        return_value="""\
              total        used        free      shared  buff/cache   available
Mem:          15658        2043       10232          12        3382       13268
Swap:          4095          39        4056
""")
    @HelperMethods.patch_clean_configuration(
        configuration={
            'limits': {
                'ram': {
                    'warning': 25,
                    'danger': 45
                }
            }
        }
    )
    def test_free_dash_m(self, _):
        """Test `free -m` output parsing for low RAM use case"""
        output_mock = MagicMock()
        RAM().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '{0}2043 MiB{1} / 15658 MiB'.format(
                Colors.GREEN_NORMAL,
                Colors.CLEAR
            )
        )

    @patch(
        'archey.entries.ram.check_output',
        return_value="""\
          total     used    free    shared  buff/cache   available
Mem:       7412     3341    1503       761        2567        3011
Swap:      7607        5    7602
""")
    @HelperMethods.patch_clean_configuration(
        configuration={
            'limits': {
                'ram': {
                    'warning': 33.3,
                    'danger': 66.7
                }
            }
        }
    )
    def test_free_dash_m_warning(self, _):
        """Test `free -m` output parsing for warning RAM use case"""
        output_mock = MagicMock()
        RAM().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '{0}3341 MiB{1} / 7412 MiB'.format(
                Colors.YELLOW_NORMAL,
                Colors.CLEAR
            )
        )

    @patch(
        'archey.entries.ram.check_output',
        return_value="""\
              total        used        free      shared  buff/cache   available
Mem:          15658       12341         624         203        2692        2807
Swap:          4095         160        3935
""")
    @HelperMethods.patch_clean_configuration(
        configuration={
            'limits': {
                'ram': {
                    'warning': 33.3,
                    'danger': 66.7
                }
            }
        }
    )
    def test_free_dash_m_danger(self, _):
        """Test `free -m` output parsing for danger RAM use case"""
        output_mock = MagicMock()
        RAM().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '{0}12341 MiB{1} / 15658 MiB'.format(
                Colors.RED_NORMAL,
                Colors.CLEAR
            )
        )

    @patch(
        'archey.entries.ram.check_output',
        side_effect=IndexError()  # `free` call will fail
    )
    @patch(
        'archey.entries.ram.open',
        mock_open(
            read_data="""\
MemTotal:        7581000 kB
MemFree:          716668 kB
MemAvailable:    3632244 kB
Buffers:          478524 kB
Cached:          2807032 kB
SwapCached:        67092 kB
Active:          3947284 kB
Inactive:        2447708 kB
Active(anon):    2268724 kB
Inactive(anon):  1106220 kB
Active(file):    1678560 kB
Inactive(file):  1341488 kB
Unevictable:         128 kB
Mlocked:             128 kB
SwapTotal:       7811068 kB
SwapFree:        7277708 kB
Dirty:               144 kB
Writeback:             0 kB
AnonPages:       3067204 kB
Mapped:           852272 kB
Shmem:            451056 kB
Slab:             314100 kB
SReclaimable:     200792 kB
SUnreclaim:       113308 kB
"""),  # Some lines have been ignored as they are useless for computations.
        create=True
    )
    def test_proc_meminfo(self, _):
        """Test `/proc/meminfo` parsing (when `free` is not available)"""
        self.assertDictEqual(
            RAM().value,
            {
                'used': 3739.296875,
                'total': 7403.3203125,
                'unit': 'MiB'
            }
        )

    @patch(
        'archey.entries.ram.check_output',
        side_effect=IndexError()  # `free` call will fail
    )
    @patch(
        'archey.entries.ram.open',
        side_effect=PermissionError(),
        create=True
    )
    @HelperMethods.patch_clean_configuration
    def test_not_detected(self, _, __):
        """Check Archey does not crash when `/proc/meminfo` is not readable"""
        ram = RAM()

        output_mock = MagicMock()
        ram.output(output_mock)

        self.assertIsNone(ram.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""Number of installed packages detection class"""

import os

from subprocess import check_output, DEVNULL, CalledProcessError

from archey.entry import Entry


PACKAGES_TOOLS = (
    {'cmd': ('apk', 'list', '--installed')},
    # As of 2020, `apt` is _very_ slow compared to `dpkg` on Debian-based distributions.
    # Additional note : `apt`'s CLI is currently not "stable" in Debian terms.
    # If `apt` happens to be preferred over `dpkg` in the future, don't forget to remove the latter.
    #{'cmd': ('apt', 'list', '-qq', '--installed')},
    {'cmd': ('dnf', 'list', 'installed'), 'skew': 1},
    {'cmd': ('dpkg', '--get-selections')},
    {'cmd': ('emerge', '-ep', 'world'), 'skew': 5},
    {'cmd': ('nix-env', '-q')},
    {'cmd': ('pacman', '-Q')},
    {'cmd': ('pkg_info', '-a')},
    {'cmd': ('pkg', '-N', 'info', '-a')},
    {'cmd': ('rpm', '-qa')},
    {'cmd': ('ls', '-l', '/var/log/packages/')},  # SlackWare.
    {'cmd': ('yum', 'list', 'installed'), 'skew': 2},
    {'cmd': ('zypper', 'search', '-i'), 'skew': 5}
)


class Packages(Entry):
    """Relies on the first found packages manager to list the installed packages"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for packages_tool in PACKAGES_TOOLS:
            try:
                results = check_output(
                    packages_tool['cmd'],
                    stderr=DEVNULL,
                    env={
                        'LANG': 'C',
                        # Alpine Linux: We have to manually propagate `PATH`.
                        #               `apk` wouldn't be found otherwise.
                        'PATH': os.getenv('PATH')
                    },
                    universal_newlines=True
                )
            except (FileNotFoundError, CalledProcessError):
                continue

            # Here we *may* use `\n` as `universal_newlines` has been set to `True`.
            if self.value:
                self.value += results.count('\n')
            else:
                self.value = results.count('\n')

            # If any, deduct output skew present due to the packages tool.
            if 'skew' in packages_tool:
                self.value -= packages_tool['skew']

            # For DPKG only, remove any not purged package.
            if packages_tool['cmd'][0] == 'dpkg':
                self.value -= results.count('deinstall')

            # Let's just loop over, in case there are multiple package managers.



"""Test module for Archey's disks usage detection module"""

import os
import unittest
from unittest.mock import call, patch, MagicMock

from archey.colors import Colors
from archey.entries.disk import Disk
from archey.test.entries import HelperMethods


class TestDiskEntry(unittest.TestCase):
    """
    Here, we mock `check_output` calls to disk utility tools.
    """
    def setUp(self):
        """We use these mocks so often, it's worth defining them here."""
        self.disk_instance_mock = HelperMethods.entry_mock(Disk)
        self.output_mock = MagicMock()

    def test_disk_get_local_filesystems(self):
        """Tests `Disk._get_local_filesystems`."""
        # This minimal `_disk_dict` contains everything this method touches.
        self.disk_instance_mock._disk_dict = {  # pylint: disable=protected-access
            '/very/good/mountpoint': {
                'device_path': '/dev/sda1'
            },
            '/mounted/here/too': {
                'device_path': '/dev/sda1'
            },
            '/other/acceptable/device/paths': {
                'device_path': '/dev/anything-really'
            },
            '/a/samba/share': {
                'device_path': '//server.local/cool_share'  # ignored - not `/dev/...`
            },
            '/linux/loop/device/one': {
                'device_path': '/dev/loop0'  # ignored - loop device
            },
            '/linux/loop/device/two': {
                'device_path': '/dev/blah/loop0'  # ignored - loop device
            },
            '/bsd/s/loop/device/one': {
                'device_path': '/dev/svnd'  # ignored - loop device
            },
            '/bsd/s/loop/device/two': {
                'device_path': '/dev/blah/svnd1'  # ignored - loop device
            },
            '/bsd/r/loop/device/one': {
                'device_path': '/dev/rvnd'  # ignored - loop device
            },
            '/bsd/r/loop/device/two': {
                'device_path': '/dev/blah/rvnd1'  # ignored - loop device
            },
            '/solaris/loop/device/one': {
                'device_path': '/dev/lofi1'  # ignored - loop device
            },
            '/solaris/loop/device/two': {
                'device_path': '/dev/blah/lofi'  # ignored - loop device
            },
            '/linux/device/mapper': {
                'device_path': '/dev/dm-1'  # ignored - device mapper
            }
        }

        result_disk_dict = Disk._get_local_filesystems(self.disk_instance_mock)  # pylint: disable=protected-access
        # Python < 3.6 doesn't guarantee dict ordering,
        # so we can't know which `/dev/sda1` mount point was used.
        self.assertEqual(
            len(result_disk_dict),
            2  # (/dev/sda1 is de-duplicated)
        )
        self.assertIn(
            '/other/acceptable/device/paths',
            result_disk_dict
        )

        # If we can now find `/dev/sda1`, then we logically must have the correct result.
        self.assertTrue(
            any(
                disk_data['device_path'] == '/dev/sda1'
                for disk_data in result_disk_dict.values()
            ),
            msg='`/dev/sda1` missing from results dict'
        )

    def test_disk_get_specified_filesystems(self):
        """Tests `Disk._get_specified_filesystems`."""
        # This minimal `_disk_dict` contains everything this method touches.
        self.disk_instance_mock._disk_dict = {  # pylint: disable=protected-access
            '/very/good/mountpoint': {
                'device_path': '/dev/sda1'
            },
            '/mounted/here/too': {
                'device_path': '/dev/sda1'
            },
            '/less/good/mountpoint': {
                'device_path': '/dev/sda2'
            },
            '/a/samba/share': {
                'device_path': '//server.local/cool_share'
            }
        }

        with self.subTest('Get all filesystems with mount points.'):
            # pylint: disable=protected-access
            self.assertDictEqual(
                Disk._get_specified_filesystems(
                    self.disk_instance_mock,
                    self.disk_instance_mock._disk_dict  # recall dicts are iterables of their keys.
                ),
                self.disk_instance_mock._disk_dict
            )
            # pylint: enable=protected-access

        with self.subTest('Get only `/dev/sda1` filesystems.'):
            result_disk_dict = Disk._get_specified_filesystems(  # pylint: disable=protected-access
                self.disk_instance_mock,
                ('/dev/sda1',)
            )

            # With Python < 3.6, dict ordering isn't guaranteed,
            # so we don't know which disk will be selected.
            self.assertEqual(len(result_disk_dict), 1)
            # As long as `device_path` is also correct, this passes.
            self.assertEqual(
                result_disk_dict[list(result_disk_dict.keys())[0]]['device_path'],
                '/dev/sda1'
            )


    @patch(
        'archey.entries.disk.check_output',
        side_effect=[
            # First `df` call succeeds.
            os.linesep.join((
                "Filesystem               1024-blocks      Used     Available Capacity Mounted on",
                "/dev/nvme0n1p2             499581952 427458276      67779164      87% /",
                "tmpfs                        8127236       292       8126944       1% /tmp",
                "/dev/nvme0n1p1                523248     35908        487340       7% /boot",
                ""
            )),
            # Second `df` call fails (emulating it not being present).
            FileNotFoundError
        ]
    )
    def test_disk_df_output_dict(self, _):
        """Test method to get `df` output as a dict by mocking calls to `check_output`."""
        self.assertDictEqual(
            Disk.get_df_output_dict(),
            {
                '/': {
                    'device_path': '/dev/nvme0n1p2',
                    'used_blocks': 427458276,
                    'total_blocks': 499581952
                },
                '/tmp': {
                    'device_path': 'tmpfs',
                    'used_blocks': 292,
                    'total_blocks': 8127236
                },
                '/boot': {
                    'device_path': '/dev/nvme0n1p1',
                    'used_blocks': 35908,
                    'total_blocks': 523248
                }
            }
        )

        with self.subTest('Missing `df` from system.'):
            self.assertDictEqual(
                Disk.get_df_output_dict(),
                {}
            )

    def test_disk_blocks_to_human_readable(self):
        """Test method to convert 1024-byte blocks to a human readable format."""
        # Each tuple is a number of blocks followed by the expected output.
        test_cases = (
            (1, '1.0 KiB'),
            (1024, '1.0 MiB'),
            (2048, '2.0 MiB'),
            (95604, '93.4 MiB'),
            (1048576, '1.0 GiB'),
            (2097152, '2.0 GiB'),
            (92156042, '87.9 GiB'),
            (1073742000, '1.0 TiB'),
            (2147484000, '2.0 TiB'),
            (458028916298, '426.6 TiB'),
            (1099512000000, '1.0 PiB'),
            (2199023000000, '2.0 PiB')  # I think we can safely stop here :)
        )
        for test_case in test_cases:
            with self.subTest(test_case[1]):
                self.assertEqual(
                    Disk._blocks_to_human_readable(test_case[0]),  # pylint: disable=protected-access
                    test_case[1]
                )

    def test_disk_output_colors(self):
        """Test `output` disk level coloring."""
        # This dict's values are tuples of used blocks, and the level's corresponding color.
        # For reference, this test uses a disk whose total block count is 100.
        levels = {
            'normal': (45.0, Colors.GREEN_NORMAL),
            'warning': (70.0, Colors.YELLOW_NORMAL),
            'danger': (95.0, Colors.RED_NORMAL)
        }
        for level, blocks_color_tuple in levels.items():
            with self.subTest(level):
                self.disk_instance_mock.value = {
                    'mount_point': {
                        'device_path': '/dev/my-cool-disk',
                        'used_blocks': blocks_color_tuple[0],
                        'total_blocks': 100
                    }
                }
                Disk.output(self.disk_instance_mock, self.output_mock)
                self.output_mock.append.assert_called_with(
                    'Disk',
                    '{color}{used} KiB{clear} / 100.0 KiB'.format(
                        color=blocks_color_tuple[1],
                        used=blocks_color_tuple[0],
                        clear=Colors.CLEAR
                    )
                )

    def test_disk_multiline_output(self):
        """Test `output`'s multi-line capability."""
        self.disk_instance_mock.value = {
            'first_mount_point': {
                'device_path': '/dev/my-cool-disk',
                'used_blocks': 10,
                'total_blocks': 10
            },
            'second_mount_point': {
                'device_path': '/dev/my-cooler-disk',
                'used_blocks': 10,
                'total_blocks': 30
            }
        }

        with self.subTest('Single-line combined output.'):
            Disk.output(self.disk_instance_mock, self.output_mock)
            self.output_mock.append.assert_called_once_with(
                'Disk',
                '{0}20.0 KiB{1} / 40.0 KiB'.format(Colors.YELLOW_NORMAL, Colors.CLEAR)
            )

        self.output_mock.reset_mock()

        with self.subTest('Multi-line output'):
            self.disk_instance_mock._configuration['disk']['combine_total'] = False  # pylint: disable=protected-access
            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk',
                        '{0}10.0 KiB{1} / 10.0 KiB'.format(Colors.RED_NORMAL, Colors.CLEAR)
                    ),
                    call(
                        'Disk',
                        '{0}10.0 KiB{1} / 30.0 KiB'.format(Colors.GREEN_NORMAL, Colors.CLEAR)
                    )
                ],
                any_order=True  # Since Python < 3.6 doesn't have definite `dict` ordering.
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (device path with entry name)'):
            self.disk_instance_mock._configuration['disk']['combine_total'] = False  # pylint: disable=protected-access
            self.disk_instance_mock._configuration['disk']['disk_labels'] = 'device_paths'  # pylint: disable=protected-access

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk (/dev/my-cool-disk)',
                        '{0}10.0 KiB{1} / 10.0 KiB'.format(Colors.RED_NORMAL, Colors.CLEAR)
                    ),
                    call(
                        'Disk (/dev/my-cooler-disk)',
                        '{0}10.0 KiB{1} / 30.0 KiB'.format(Colors.GREEN_NORMAL, Colors.CLEAR)
                    )
                ],
                any_order=True  # Since Python < 3.6 doesn't have definite `dict` ordering.
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (mount points without entry name)'):
            self.disk_instance_mock._configuration['disk']['combine_total'] = False  # pylint: disable=protected-access
            self.disk_instance_mock._configuration['disk']['disk_labels'] = 'mount_points'  # pylint: disable=protected-access
            self.disk_instance_mock._configuration['disk']['hide_entry_name'] = True  # pylint: disable=protected-access

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        '(first_mount_point)',
                        '{0}10.0 KiB{1} / 10.0 KiB'.format(Colors.RED_NORMAL, Colors.CLEAR)
                    ),
                    call(
                        '(second_mount_point)',
                        '{0}10.0 KiB{1} / 30.0 KiB'.format(Colors.GREEN_NORMAL, Colors.CLEAR)
                    )
                ],
                any_order=True  # Since Python < 3.6 doesn't have definite `dict` ordering.
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (without disk label nor entry name)'):
            self.disk_instance_mock._configuration['disk']['combine_total'] = False  # pylint: disable=protected-access
            self.disk_instance_mock._configuration['disk']['disk_labels'] = False  # pylint: disable=protected-access
            # `hide_entry_name` is being ignored as `disk_labels` evaluates to "falsy" too.
            self.disk_instance_mock._configuration['disk']['hide_entry_name'] = True  # pylint: disable=protected-access

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk',
                        '{0}10.0 KiB{1} / 10.0 KiB'.format(Colors.RED_NORMAL, Colors.CLEAR)
                    ),
                    call(
                        'Disk',
                        '{0}10.0 KiB{1} / 30.0 KiB'.format(Colors.GREEN_NORMAL, Colors.CLEAR)
                    )
                ],
                any_order=True  # Since Python < 3.6 doesn't have definite `dict` ordering.
            )


if __name__ == '__main__':
    unittest.main()



"""Shell detection class"""

import os

from subprocess import CalledProcessError, check_output

from archey.entry import Entry


class Shell(Entry):
    """
    Simple shell path detection based either on the `SHELL` environment variable or
    the local administrative database.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = os.getenv('SHELL')
        if not self.value:
            try:
                self.value = check_output(
                    ['getent', 'passwd', str(os.getuid())],
                    universal_newlines=True
                ).rstrip().split(':')[-1]
            except CalledProcessError:
                # Where does this user come from ?
                pass



"""Hardware model information detection class"""

import os
import re

from subprocess import CalledProcessError, DEVNULL, check_output

from archey.entry import Entry


class Model(Entry):
    """Uses multiple methods to retrieve some information about the host hardware"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = \
            self._fetch_virtual_env_info() \
            or self._fetch_product_info() \
            or self._fetch_rasperry_pi_revision() \
            or self._fetch_android_device_model() \

    def _fetch_virtual_env_info(self):
        """
        Relying on some system tools, tries to gather some details about hypervisor.
        When available, relies on systemd.
        When run as root, `virt-what` and/or `dmidecode` may be called.
        """
        environment, product_name = None, None

        try:
            environment = check_output(
                ['systemd-detect-virt'],
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except CalledProcessError:
            # Not a virtual environment.
            return None
        except FileNotFoundError:
            pass

        # When run as root, let's ask `virt-what` and/or `dmidecode`.
        if os.getuid() == 0:
            # We couldn't retrieve any information from `systemd-detect-virt`.
            if not environment:
                try:
                    environment = ', '.join(
                        check_output(
                            ['virt-what'],
                            stderr=DEVNULL, universal_newlines=True
                        ).splitlines()
                    )
                except (FileNotFoundError, CalledProcessError):
                    pass

                # Definitely not a virtual environment.
                if not environment:
                    return None

            try:
                # Sometimes we may gather info added by hosting service provider this way.
                product_name = check_output(
                    ['dmidecode', '-s', 'system-product-name'],
                    stderr=DEVNULL, universal_newlines=True
                ).rstrip()
            except (FileNotFoundError, CalledProcessError):
                pass
        elif not environment:
            # No detection tool is available...
            return None

        # If we got there with some info, this _should_ be a virtual environment.
        return '{0} ({1})'.format(
            product_name or self._configuration.get('default_strings')['virtual_environment'],
            environment
        )

    @staticmethod
    def _fetch_product_info():
        """Tries to open specific Linux files, looking for hardware product name and version"""
        try:
            with open('/sys/devices/virtual/dmi/id/product_name') as f_product_name:
                product_name = f_product_name.read().rstrip()
        except FileNotFoundError:
            return None

        try:
            with open('/sys/devices/virtual/dmi/id/product_version') as f_product_version:
                product_version = f_product_version.read().rstrip()
        except FileNotFoundError:
            product_version = None

        if not product_version:
            return product_name

        return "{} {}".format(product_name, product_version)

    @staticmethod
    def _fetch_rasperry_pi_revision():
        """Tries to retrieve 'Hardware' and 'Revision IDs' from `/proc/cpuinfo`"""
        try:
            with open('/proc/cpuinfo') as f_cpu_info:
                cpu_info = f_cpu_info.read()
        except (PermissionError, FileNotFoundError):
            return None

        # If the output contains 'Hardware' and 'Revision'...
        hardware = re.search('(?<=Hardware\t: ).*', cpu_info)
        revision = re.search('(?<=Revision\t: ).*', cpu_info)
        if hardware and revision:
            # ... let's set a pretty info string with these data
            return 'Raspberry Pi {0} (Rev. {1})'.format(
                hardware.group(0),
                revision.group(0)
            )

        return None

    @staticmethod
    def _fetch_android_device_model():
        """Tries to retrieve `brand` and `model` device properties on Android platforms"""
        try:
            brand = check_output(
                ['getprop', 'ro.product.brand'],
                universal_newlines=True
            ).rstrip()
            model = check_output(
                ['getprop', 'ro.product.model'],
                universal_newlines=True
            ).rstrip()
        except FileNotFoundError:
            return None

        return '{0} ({1})'.format(brand, model)



"""Test module for Archey's distribution detection module"""

import unittest
from unittest.mock import MagicMock, patch

from archey.entries.distro import Distro
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestDistroEntry(unittest.TestCase):
    """`Distro` entry simple test cases"""
    @patch(
        'archey.entries.distro.Distributions.get_distro_name',
        return_value='NAME VERSION (CODENAME)'
    )
    @patch(
        'archey.entries.distro.Distro._fetch_architecture',
        return_value='ARCHITECTURE'
    )
    def test_init(self, _, __):
        """Test `Distro` instantiation"""
        self.assertDictEqual(
            Distro().value,
            {
                'name': 'NAME VERSION (CODENAME)',
                'arch': 'ARCHITECTURE'
            }
        )

    @patch(
        'archey.entries.distro.check_output',
        return_value='x86_64\n'  # Imitate `uname` output on AMD64.
    )
    def test_fetch_architecture(self, _):
        """Test `_fetch_architecture` static method"""
        self.assertEqual(
            Distro._fetch_architecture(),  # pylint: disable=protected-access
            'x86_64'
        )

    @patch(
        'archey.entries.distro.check_output',
        return_value='10\n'  # Imitate `getprop` output on Android 10.
    )
    def test_fetch_android_release(self, _):
        """Test `_fetch_android_release` static method"""
        self.assertEqual(
            Distro._fetch_android_release(),  # pylint: disable=protected-access
            'Android 10'
        )

    @patch(
        'archey.entries.distro.Distributions.get_distro_name',
        return_value=None  # Soft-failing : No _pretty_ distribution name found...
    )
    @patch(
        'archey.entries.distro.Distro._fetch_architecture',
        return_value='ARCHITECTURE'
    )
    @patch(
        'archey.entries.distro.Distro._fetch_android_release',
        return_value=None  # Not an Android device either...
    )
    @HelperMethods.patch_clean_configuration
    def test_unknown_distro_output(self, _, __, ___):
        """Test for `distro` and `uname` outputs concatenation"""
        distro = Distro()

        output_mock = MagicMock()
        distro.output(output_mock)

        self.assertDictEqual(
            distro.value,
            {
                'name': None,
                'arch': 'ARCHITECTURE'
            }
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected'] + ' [ARCHITECTURE]'
        )


if __name__ == '__main__':
    unittest.main()



"""Desktop environment detection class"""

import os

from archey.entry import Entry
from archey.processes import Processes


DE_DICT = {
    'cinnamon': 'Cinnamon',
    'dde-dock': 'Deepin',
    'fur-box-session': 'Fur Box',
    'gnome-session': 'GNOME',
    'gnome-shell': 'GNOME',
    'ksmserver': 'KDE',
    'lxqt-session': 'LXQt',
    'lxsession': 'LXDE',
    'mate-session': 'MATE',
    'xfce4-session': 'Xfce'
}


class DesktopEnvironment(Entry):
    """
    Just iterate over running processes to find a known-entry.
    If not, rely on the `XDG_CURRENT_DESKTOP` environment variable.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        processes = Processes().list
        for de_id, de_name in DE_DICT.items():
            if de_id in processes:
                self.value = de_name
                break
        else:
            # Let's rely on an environment variable if the loop above didn't `break`.
            self.value = os.getenv('XDG_CURRENT_DESKTOP')



"""Uptime detection class"""

import re

from subprocess import check_output

import sys
import time

from datetime import timedelta

from archey.entry import Entry


class Uptime(Entry):
    """Returns a pretty-formatted string representing the host uptime"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        uptime_seconds = int(self._get_uptime_delta().total_seconds())

        days, uptime_seconds = divmod(uptime_seconds, 86400)
        hours, uptime_seconds = divmod(uptime_seconds, 3600)
        minutes, seconds = divmod(uptime_seconds, 60)

        self.value = {
            'days': days,
            'hours': hours,
            'minutes': minutes,
            'seconds': seconds
        }

    def _get_uptime_delta(self):
        """
        Returns a `datetime.timedelta` instance containing the machine uptime.
        Tries a variety of methods, increasing compatibility for a wide range of systems.
        """
        # Try the /proc/uptime file
        try:
            return self._proc_file_uptime()
        except (PermissionError, FileNotFoundError):
            # Can't read /proc/uptime.
            # Not GNU/Linux ? Limited permissions ?
            pass

        # Try the python `time` module clocks
        try:
            return self._clock_uptime()
        except RuntimeError:
            # Probably Python <3.7, or not *NIX
            pass

        # FUTURE: Windows support could be added here with the `wmi` or `pywin32` module.

        # Fall back to the `uptime` command.
        return self._parse_uptime_cmd()

    @staticmethod
    def _proc_file_uptime():
        """Tries to get uptime using the `/proc/uptime` file"""
        with open('/proc/uptime') as f_uptime:
            return timedelta(
                seconds=float(f_uptime.read().split()[0])
            )

    @staticmethod
    def _clock_uptime():
        """Tries to get uptime using the clocks from the Python `time` module"""
        # Try: Linux uptime clock, macOS uptime clock, BSD uptime clock.
        for clock in ('CLOCK_BOOTTIME', 'CLOCK_UPTIME_RAW', 'CLOCK_UPTIME'):
            try:
                return timedelta(
                    seconds=time.clock_gettime(getattr(time, clock))
                )
            except AttributeError:
                pass

        # Probably Python <3.7, or just not one of the above OSes
        raise RuntimeError

    @staticmethod
    def _parse_uptime_cmd():
        """Tries to get uptime by parsing the `uptime` command"""
        try:
            uptime_output = check_output('uptime', env={'LANG': 'C'})
        except FileNotFoundError:
            # No `uptime` command.
            # Since `procps` is a dependency (which provides `uptime`) we can just exit here.
            # Note: We shouldn't get there as `Processes` depends on `procps` beforehand.
            sys.exit("Please, install first `procps` (or `procps-ng`) on your system.")

        # Unfortunately the output is not designed to be machine-readable...
        uptime_match = re.search(
            r"""
            up\s+?             # match the `up` preceding the uptime (anchor the start of the regex)
            (?:                # non-capture group for days section.
               (?P<days>       # 'days' named capture group, captures the days digits.
                  \d+?
               )
               \s+?            # match whitespace,
               days?           #   'day' or 'days',
               [,\s]+?         #   then a comma (if present) followed by more whitespace.
            )?                 # match the days non-capture group 0 or 1 times.
            (?:                # non-capture group for hours & minutes section.
               (?:             # non-capture group for just hours section.
                  (?P<hours>   # 'hours' named capture group, captures the hours digits.
                     \d+?
                  )
                  (?:          # non-capture group for hours:minutes colon or 'hrs' text...
                     :         #   i.e. hours followed by either a single colon
                     |         #   OR
                     \s+?      #   1 or more whitespace chars non-greedily,
                     hrs?      #   followed by 'hr' or 'hrs'.
                  )
               )?              # match the hours non-capture group 0 or 1 times.
               (?:             # non-capture group for minutes section.
                  (?P<minutes> # 'minutes' named capture group, captures the minutes digits.
                     \d+?
                  )
                  (?:          # non-capture group for 'min' or 'mins' text.
                     \s+?      # match whitespace,
                     mins?     #   followed by 'min' or 'mins'.
                  )?           # match the text 0 or 1 times (0 times is for the hh:mm format case).
                  (?!          # negative lookahead group
                     \d+       #   this prevents matching seconds digits as minutes...
                     \s+?      #   since we only non-greedily match minutes digits earlier.
                     secs?     #   here's the part that matches the 'sec' or 'secs' text.
                  )            # the minutes group is discarded if this lookahead matches!
               )?              # match the minutes non-capture group 0 or 1 times.
            )?                 # match the entire hours & minutes non-capture group 0 or 1 times.
            (?:                # non-capture group for seconds.
               (?P<seconds>    # 'seconds' named capture group, captures the seconds digits.
                  \d+?
               )
               \s+?            # match whitespace,
               secs?           #  then 'sec' or 'secs'.
            )?                 # match the seconds non-capture group 0 or 1 times.
            [,\s]*?            # after the groups, match a comma and/or whitespace 0 or more times,
            \d+?               #   one or more digits for the user count,
            \s+?               #   whitespace between the user count and the text 'user',
            user               #   and the text 'user' (to anchor the end of the expression).
            """,
            uptime_output,
            re.VERBOSE
        )

        # Only `days`, `hours`, `minutes` or `seconds` could have been captured.
        # `timedelta` directly accepts them as arguments.
        uptime_args = uptime_match.groupdict()
        return timedelta(
            days=int(uptime_args.get('days') or 0),
            hours=int(uptime_args.get('hours') or 0),
            minutes=int(uptime_args.get('minutes') or 0),
            seconds=int(uptime_args.get('seconds') or 0)
        )


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting the uptime to a string."""
        days = self.value['days']
        hours = self.value['hours']
        minutes = self.value['minutes']

        uptime = ''
        if days:
            uptime += str(days) + ' day'
            if days > 1:
                uptime += 's'

            if hours or minutes:
                if bool(hours) != bool(minutes):
                    uptime += ' and '
                else:
                    uptime += ', '

        if hours:
            uptime += str(hours) + ' hour'
            if hours > 1:
                uptime += 's'

            if minutes:
                uptime += ' and '

        if minutes:
            uptime += str(minutes) + ' minute'
            if minutes > 1:
                uptime += 's'
        elif not days and not hours:
            uptime = '< 1 minute'

        output.append(self.name, uptime)



"""Test module for Archey's temperature detection module"""

import os
import tempfile

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, patch


from archey.entries.temperature import Temperature
from archey.test.entries import HelperMethods


class TestTemperatureEntry(unittest.TestCase):
    """
    Based on `sensors`, `vcgencmd` and thermal files, this module verifies temperature computations.
    """

    def setUp(self):
        # We'll store there filenames of some temp files mocking those under
        #  `/sys/class/thermal/thermal_zone*/temp`
        self._temp_files = []
        for temperature in [  # Fake temperatures
                b'50000',
                b'0',
                b'40000',
                b'50000'
            ]:
            file = tempfile.NamedTemporaryFile(delete=False)
            file.write(temperature)
            file.seek(0)
            self._temp_files.append(file)

    def tearDown(self):
        for file in self._temp_files:
            file.close()
            os.remove(file.name)

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),
            'temp=42.8\'C\n'
        ]
    )
    @patch(
        'archey.entries.temperature.iglob',
        return_value=[]  # No temperature from file will be retrieved
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': False,
                'char_before_unit': ' '
            }
        }
    )
    def test_vcgencmd_only_no_max(self, _, __):
        """
        Test for `vcgencmd` output only (no sensor files).
        Only one value is retrieved, so no maximum should be displayed (see #39).
        """
        temperature = Temperature()

        output_mock = MagicMock()
        temperature.output(output_mock)

        self.assertDictEqual(
            temperature.value,
            {
                'temperature': 42.8,
                'max_temperature': 42.8,
                'char_before_unit': ' ',
                'unit': 'C'
            }
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '42.8 C'
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),
            'temp=40.0\'C\n'
        ]
    )
    @patch('archey.entries.temperature.iglob')
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': False,
                'char_before_unit': ' '
            }
        }
    )
    def test_vcgencmd_and_files(self, iglob_mock, _):
        """Tests `vcgencmd` output AND sensor files"""
        iglob_mock.return_value = iter([file.name for file in self._temp_files])
        self.assertDictEqual(
            Temperature().value,
            {
                'temperature': 45.0,
                'max_temperature': 50.0,
                'char_before_unit': ' ',
                'unit': 'C'
            }
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),  # No temperature from `sensors` call
            FileNotFoundError()   # No temperature from `vcgencmd` call
        ]
    )
    @patch('archey.entries.temperature.iglob')
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': True,
                'char_before_unit': '@'
            }
        }
    )
    def test_files_only_in_fahrenheit(self, iglob_mock, _):
        """Test sensor files only, Fahrenheit (naive) conversion and special degree character"""
        iglob_mock.return_value = iter([file.name for file in self._temp_files])
        self.assertDictEqual(
            Temperature().value,
            {
                'temperature': 116.0,      # 46.7 degrees C in Fahrenheit.
                'max_temperature': 122.0,  # 50 degrees C in Fahrenheit
                'char_before_unit': '@',
                'unit': 'F'
            }
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),  # No temperature from `sensors` call.
            FileNotFoundError()   # No temperature from `vcgencmd` call.
        ]
    )
    @patch(
        'archey.entries.temperature.iglob',
        return_value=[]  # No temperature from file will be retrieved.
    )
    @HelperMethods.patch_clean_configuration(
        configuration={'temperature': {'sensors_chipsets': []}}
    )
    def test_no_output(self, _, __):
        """Test when no value could be retrieved (anyhow)"""
        self.assertIsNone(Temperature().value)

    @patch(
        'archey.entries.temperature.check_output',  # Mock the `sensors` call.
        side_effect=[
            """\
{
   "who-cares-about":{
      "temp1":{
         "temp1_input": 45.000,
         "temp1_crit": 128.000
      },
      "temp2":{
         "temp2_input": 0.000,
         "temp2_crit": 128.000
      },
      "temp3":{
         "temp3_input": 38.000,
         "temp3_crit": 128.000
      },
      "temp4":{
         "temp4_input": 39.000,
         "temp4_crit": 128.000
      },
      "temp5":{
         "temp5_input": 0.000,
         "temp5_crit": 128.000
      },
      "temp6":{
         "temp6_input": 114.000,
         "temp6_crit": 128.000
      }
   },
   "the-chipsets-names":{
      "what-are":{
         "temp1_input": 45.000,
         "temp1_max": 100.000,
         "temp1_crit": 100.000,
         "temp1_crit_alarm": 0.000
      },
      "those":{
         "temp2_input": 43.000,
         "temp2_max": 100.000,
         "temp2_crit": 100.000,
         "temp2_crit_alarm": 0.000
      },
      "identifiers":{
         "temp3_input": 44.000,
         "temp3_max": 100.000,
         "temp3_crit": 100.000,
         "temp3_crit_alarm": 0.000
      }
   },
   "crap-a-fan-chip":{
      "fan1":{
         "fan1_input": 3386.000
      }
   }
}
""",
            FileNotFoundError()  # No temperature from `vcgencmd` call.
        ]
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': True,
                'char_before_unit': ' '
            }
        }
    )
    def test_sensors_only_in_fahrenheit(self, _):
        """Test computations around `sensors` output and Fahrenheit (naive) conversion"""
        self.assertDictEqual(
            Temperature().value,
            {
                'temperature': 126.6,      # (52.6 C in F)
                'max_temperature': 237.2,  # (114.0 C in F)
                'char_before_unit': ' ',
                'unit': 'F'
            }
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            CalledProcessError(1, 'sensors'),  # `sensors` will hard fail.
            FileNotFoundError()                # No temperature from `vcgencmd` call
        ]
    )
    @patch('archey.entries.temperature.iglob')
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': False,
                'char_before_unit': 'o'
            }
        }
    )
    def test_sensors_error_1(self, iglob_mock, _):
        """Test `sensors` (hard) failure handling and polling from files in Celsius"""
        iglob_mock.return_value = iter([file.name for file in self._temp_files])

        temperature = Temperature()

        output_mock = MagicMock()
        temperature.output(output_mock)

        self.assertDictEqual(
            temperature.value,
            {
                'temperature': 46.7,
                'max_temperature': 50.0,
                'char_before_unit': 'o',
                'unit': 'C'
            }
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '46.7oC (Max. 50.0oC)'
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[  # JSON decoding from `sensors` will fail..
            """\
{
    "Is this JSON valid ?": [
        "You", "should", "look", "twice.",
    ]
}
""",
            FileNotFoundError()  # No temperature from `vcgencmd` call
        ]
    )
    @patch('archey.entries.temperature.iglob')
    @HelperMethods.patch_clean_configuration(
        configuration={
            'temperature': {
                'sensors_chipsets': [],
                'use_fahrenheit': False,
                'char_before_unit': 'o'
            }
        }
    )
    def test_sensors_error_2(self, iglob_mock, _):
        """Test `sensors` (hard) failure handling and polling from files in Celsius"""
        iglob_mock.return_value = iter([file.name for file in self._temp_files])
        self.assertDictEqual(
            Temperature().value,
            {
                'temperature': 46.7,
                'max_temperature': 50.0,
                'char_before_unit': 'o',
                'unit': 'C'
            }
        )

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),  # No temperature from `sensors` call.
            FileNotFoundError()   # No temperature from `vcgencmd` call.
        ]
    )
    @patch(
        'archey.entries.temperature.iglob',
        return_value=[]  # No temperature from file will be retrieved.
    )
    @HelperMethods.patch_clean_configuration
    def test_celsius_to_fahrenheit_conversion(self, _, __):
        """Simple tests for the `_convert_to_fahrenheit` static method"""
        test_conversion_cases = [
            (-273.15, -459.67),
            (0.0, 32.0),
            (21.0, 69.8),
            (37.0, 98.6),
            (100.0, 212.0)
        ]

        for celsius_value, expected_fahrenheit in test_conversion_cases:
            self.assertAlmostEqual(
                Temperature._convert_to_fahrenheit(celsius_value),  # pylint: disable=protected-access
                expected_fahrenheit
            )


if __name__ == '__main__':
    unittest.main()



"""`archey.test.entries` module initialization file"""

from copy import deepcopy
from functools import wraps

import unittest
from unittest.mock import MagicMock, patch

from archey.configuration import Configuration
from archey.constants import DEFAULT_CONFIG
from archey.entry import Entry


class HelperMethods:
    """
    This class contains helper methods we commonly use in our entry unit tests.
    We kindly borrow `update_recursive` class method from `Configuration` to DRY its implementation.
    """
    @staticmethod
    def entry_mock(entry, configuration=None):
        """
        Creates a placeholder "instance" of the entry class passed, with a clean default
        `_configuration` which is optionally updated by `configuration`.

        It can be used to very cleanly unit-test instance methods of a class,
        by passing it in (after setting appropriate attributes).

        The attributes defined are not instance attributes, however since this isn't
        technically an instance, they are used in place of the respective instance attributes.
        """
        # We spec to the entry so non-existent methods can't be called...
        # ...and wrap it, to inherit its methods.
        instance_mock = MagicMock(spec=entry, wraps=entry)
        # These instance-attributes are quite important, so let's mimic them.
        instance_mock.name = str(entry.__name__)
        instance_mock.value = None  # (entry default)

        # Let's initially give the entry configuration the defaults.
        # We deep-copy `DEFAULT_CONFIG` to prevent its mutation.
        entry_configuration = deepcopy(DEFAULT_CONFIG)
        # Then, let's merge in `configuration` recursively.
        Configuration.update_recursive(entry_configuration, (configuration or {}))
        # Finally, replaces the internal (and private!) `_configuration` attribute by...
        # ... the corresponding configuration object.
        setattr(instance_mock, '_configuration', entry_configuration)

        return instance_mock

    @staticmethod
    def patch_clean_configuration(method_definition=None, *, configuration=None):
        """
        Decorator for an entry test definition, which sets the entry's `_configuration` attribute to
        the Archey defaults, optionally updated with `configuration`.
        """
        # Let's initially give the entry configuration the defaults.
        # We deep-copy `DEFAULT_CONFIG` to prevent its mutation.
        entry_configuration = deepcopy(DEFAULT_CONFIG)
        # Then, let's merge in `configuration` recursively.
        Configuration.update_recursive(entry_configuration, (configuration or {}))

        def decorator_patch_clean_configuration(method):
            @wraps(method)
            def wrapper_patch_clean_configuration(*args, **kwargs):
                with patch('archey.entry.Configuration', autospec=True) as config_instance_mock:
                    # Mock "publicly" used methods.
                    config_instance_mock().get = entry_configuration.get
                    config_instance_mock().__iter__ = lambda _: iter(entry_configuration.items())
                    return method(*args, **kwargs)

            return wrapper_patch_clean_configuration

        if method_definition is None:
            return decorator_patch_clean_configuration

        return decorator_patch_clean_configuration(method_definition)


class TestHelperMethods(unittest.TestCase, HelperMethods):
    """This class implements test cases for our helper methods."""
    class _SimpleEntry(Entry):
        """A simple (sub-)class inheriting from `Entry` to use in testing."""
        def __init__(self):
            super().__init__()
            self.name = 'meaning_of_life'
            self.value = self.meaning_of_life()

        def my_name(self):
            """Returns the entry name."""
            return self.name

        @staticmethod
        def meaning_of_life():
            """Widely debated..."""
            return 42

    def test_entry_mock_defaults(self):
        """Test `entry_mock`s default attributes."""
        simple_mock_instance = self.entry_mock(TestHelperMethods._SimpleEntry)
        self.assertEqual(simple_mock_instance.name, '_SimpleEntry')
        self.assertIsNone(simple_mock_instance.value)
        self.assertDictEqual(simple_mock_instance._configuration, DEFAULT_CONFIG)  # pylint: disable=protected-access

    def test_entry_mock_spec(self):
        """Test `entry_mock`s speccing."""
        simple_mock_instance = self.entry_mock(TestHelperMethods._SimpleEntry)
        with self.assertRaises(AttributeError):
            # We shouldn't be able to call methods that don't exist...
            simple_mock_instance.not_a_method()
            # ...or get attributes that don't exist.
            _ = simple_mock_instance.not_an_attribute
        # However we _should_ be able to set an attribute then use it.
        simple_mock_instance.is_simple = True
        self.assertTrue(simple_mock_instance.is_simple)

    def test_entry_mock_wrap(self):
        """Test `entry_mock`s class wrapping."""
        simple_mock_instance = self.entry_mock(TestHelperMethods._SimpleEntry)
        self.assertEqual(simple_mock_instance.meaning_of_life(), 42)
        simple_mock_instance.name = 'A simple entry!'
        self.assertEqual(simple_mock_instance.my_name(simple_mock_instance), 'A simple entry!')

    @patch.dict(
        DEFAULT_CONFIG,
        values={
            'a_key': 'a_value',
            'a_dict': {
                'key_1': 1,
                'key_2': 2
            }
        },
        clear=True
    )
    def test_entry_mock_configuration_setting(self):
        """Test `entry_mock`s configuration setting."""
        configuration_dict = {
            'another_key': 'another_value',  # Adding a key-value pair.
            'a_dict': {
                'key_1': 10  # Updating a nested key-value pair.
            }
        }
        simple_mock_instance = self.entry_mock(TestHelperMethods._SimpleEntry, configuration_dict)
        self.assertDictEqual(
            simple_mock_instance._configuration,  # pylint: disable=protected-access
            {
                'a_key': 'a_value',
                'another_key': 'another_value',
                'a_dict': {
                    'key_1': 10,
                    'key_2': 2
                }
            }
        )

    def test_patch_clean_configuration_defaults(self):
        """Test the default configuration-setting of `patch_clean_configuration."""
        @HelperMethods.patch_clean_configuration
        def test(self):
            simple_entry = TestHelperMethods._SimpleEntry()
            self.assertDictEqual(
                dict(simple_entry._configuration),  # pylint: disable=protected-access
                DEFAULT_CONFIG
            )

        test(self)

    @patch.dict(
        DEFAULT_CONFIG,
        values={
            'a_key': 'a_value',
            'a_dict': {
                'key_1': 1,
                'key_2': 2
            }
        },
        clear=True
    )
    def test_patch_clean_configuration_setting(self):
        """Test `patch_clean_configuration`s configuration setting."""
        configuration_dict = {
            'another_key': 'another_value',  # Adding a key-value pair.
            'a_dict': {
                'key_1': 10  # Updating a nested key-value pair.
            }
        }
        @HelperMethods.patch_clean_configuration(
            configuration=configuration_dict
        )
        def test(self):
            simple_entry = TestHelperMethods._SimpleEntry()
            self.assertDictEqual(
                dict(simple_entry._configuration),  # pylint: disable=protected-access
                {
                    'a_key': 'a_value',
                    'another_key': 'another_value',
                    'a_dict': {
                        'key_1': 10,
                        'key_2': 2
                    }
                }
            )

        test(self)



"""Temperature detection class"""

import json
import re

from glob import iglob
from subprocess import check_output, DEVNULL, CalledProcessError

from archey.entry import Entry


class Temperature(Entry):
    """
    Tries to compute an average temperature from `sensors` (LM-Sensors).
    If not available, falls back on system thermal zones files.
    On Raspberry devices, retrieves temperature from the `vcgencmd` binary.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._temps = []

        # Tries `sensors` at first.
        self._run_sensors(self._configuration.get('temperature')['sensors_chipsets'])

        # On error (list still empty), checks for system thermal zones files.
        if not self._temps:
            self._poll_thermal_zones()

        # Tries `vcgencmd` for Raspberry devices.
        self._run_vcgencmd()

        # No value could be fetched...
        if not self._temps:
            return

        # Let's DRY some constants once.
        use_fahrenheit = self._configuration.get('temperature')['use_fahrenheit']
        char_before_unit = self._configuration.get('temperature')['char_before_unit']

        # Conversion to Fahrenheit if needed.
        if use_fahrenheit:
            self._temps = list(map(self._convert_to_fahrenheit, self._temps))

        # Final average and maximum computations.
        self.value = {
            'temperature': float(round(sum(self._temps) / len(self._temps), 1)),
            'max_temperature': float(round(max(self._temps), 1)),
            'char_before_unit': char_before_unit,
            'unit': ('F' if use_fahrenheit else 'C')
        }


    def _run_sensors(self, whitelisted_chips):
        # Uses the `sensors` program (from LM-Sensors) to interrogate thermal chip-sets.
        try:
            sensors_output = check_output(
                ['sensors', '-A', '-j'] + whitelisted_chips,
                universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            return

        try:
            sensors_data = json.loads(sensors_output)
        # For backward compatibility with Python versions prior to 3.5.0
        #   we use `ValueError` instead of `json.JSONDecodeError`.
        except ValueError:
            return

        # Iterates over the chip-sets outputs to filter interesting values.
        for features in sensors_data.values():
            for subfeatures in features.values():
                for name, value in subfeatures.items():
                    # These conditions check whether this sub-feature value is a correct
                    #  temperature, as :
                    # * It might be an input fan speed (from a control chip) ;
                    # * Some chips/adapters might return null temperatures.
                    if value != 0.0 and re.match(r"temp\d_input", name):
                        self._temps.append(value)
                        # There is only one `temp*_input` field, let's stop the current iteration.
                        break

    def _poll_thermal_zones(self):
        # We just check for values within files present in the path below.
        for thermal_file in iglob('/sys/class/thermal/thermal_zone*/temp'):
            with open(thermal_file) as file:
                try:
                    temp = float(file.read().strip()) / 1000
                except OSError:
                    continue

                if temp != 0.0:
                    self._temps.append(temp)

    def _run_vcgencmd(self):
        # Let's try to retrieve a value from the Broadcom chip on Raspberry.
        try:
            vcgencmd_output = check_output(
                ['/opt/vc/bin/vcgencmd', 'measure_temp'],
                stderr=DEVNULL, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            return

        self._temps.append(
            float(
                re.search(
                    r'\d+\.\d+',
                    vcgencmd_output
                ).group(0)
            )
        )

    @staticmethod
    def _convert_to_fahrenheit(temp):
        """Simple Celsius to Fahrenheit conversion method"""
        return temp * (9 / 5) + 32


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting with units."""
        if not self.value:
            # Fall back on the default behavior if no temperatures were detected.
            super().output(output)
            return

        # DRY some constants
        char_before_unit = self.value['char_before_unit']
        unit = self.value['unit']

        entry_text = '{0}{1}{2}'.format(
            self.value['temperature'],
            char_before_unit,
            unit
        )
        # When there are multiple input sources, show the hottest value.
        if len(self._temps) > 1:
            entry_text = '{0} (Max. {1}{2}{3})'.format(
                entry_text,
                self.value['max_temperature'],
                char_before_unit,
                unit
            )

        output.append(self.name, entry_text)



"""Colors enumeration definition"""

from bisect import bisect
from enum import Enum

import os
import re


# REGEXP compiled pattern matching ANSI/ECMA-48 color escape codes.
ANSI_ECMA_REGEXP = re.compile(r'\x1b\[\d+(?:(?:;\d+)+)?m')

# Let's honor `NO_COLOR` if set.
# See <https://no-color.org/>.
NO_COLOR = ('NO_COLOR' in os.environ)


class Colors(Enum):
    """
    ANSI terminal colors enumeration.
    Supports an arbitrary number of display attributes.

    See <http://www.termsys.demon.co.uk/vtansi.htm#colors>.
    """
    CLEAR = (0,)
    RED_NORMAL = (0, 31)
    RED_BRIGHT = (1, 31)
    GREEN_NORMAL = (0, 32)
    GREEN_BRIGHT = (1, 32)
    YELLOW_NORMAL = (0, 33)
    YELLOW_BRIGHT = (1, 33)
    BLUE_NORMAL = (0, 34)
    BLUE_BRIGHT = (1, 34)
    MAGENTA_NORMAL = (0, 35)
    MAGENTA_BRIGHT = (1, 35)
    CYAN_NORMAL = (0, 36)
    CYAN_BRIGHT = (1, 36)
    WHITE_NORMAL = (0, 37)
    WHITE_BRIGHT = (1, 37)

    def __str__(self):
        return self.escape_code_from_attrs(
            ';'.join(map(str, self.value))
        )

    @staticmethod
    def escape_code_from_attrs(display_attrs):
        """
        Build and return an ANSI/ECMA-48 escape code string from passed display attributes.
        """
        if NO_COLOR:
            return ''

        return '\x1b[{}m'.format(display_attrs)

    @staticmethod
    def get_level_color(value, yellow_bpt, red_bpt):
        """Returns the best level color according to `value` compared to `{yellow,red}_bpt`"""
        level_colors = (Colors.GREEN_NORMAL, Colors.YELLOW_NORMAL, Colors.RED_NORMAL)

        return level_colors[bisect((yellow_bpt, red_bpt), value)]

    @staticmethod
    def remove_colors(string):
        """Simple DRY method to remove any ANSI/ECMA-48 color escape code from passed `string`"""
        return ANSI_ECMA_REGEXP.sub('', string)



"""
Output class file.
It supports entries lazy-insertion, logo detection, and final printing.
"""

from textwrap import TextWrapper
from shutil import get_terminal_size

import os
import sys

from archey.api import API
from archey.colors import ANSI_ECMA_REGEXP, Colors
from archey.constants import COLORS_DICT, LOGOS_DICT
from archey.configuration import Configuration
from archey.distributions import Distributions
from archey.logos import get_logo_width


class Output:
    """
    This is the object handling output entries populating.
    It also handles the logo choice based on some system detections.
    """
    def __init__(self, **kwargs):
        # Fetches passed arguments.
        self._format_to_json = kwargs.get('format_to_json')

        try:
            # If set, force the distribution to `preferred_distribution` argument.
            self._distribution = Distributions(kwargs.get('preferred_distribution'))
        except ValueError:
            # If not (or unknown), run distribution detection.
            self._distribution = Distributions.run_detection()

        # Fetch the colors palette related to this distribution.
        self._colors_palette = COLORS_DICT[self._distribution]

        # If `os-release`'s `ANSI_COLOR` option is set, honor it.
        ansi_color = Distributions.get_ansi_color()
        if ansi_color and Configuration().get('colors_palette')['honor_ansi_color']:
            # Replace each Archey integrated colors by `ANSI_COLOR`.
            self._colors_palette = len(self._colors_palette) * \
                [Colors.escape_code_from_attrs(ansi_color)]

        # Each entry will be added to this list
        self._entries = []
        # Each class output will be added in the list below afterwards
        self._results = []

    def add_entry(self, module):
        """Append an entry to the list of entries to output"""
        self._entries.append(module)

    def append(self, key, value):
        """Append a pre-formatted entry to the final output content"""
        self._results.append(
            '{color}{key}:{clear} {value}'.format(
                color=self._colors_palette[0],
                key=key,
                clear=Colors.CLEAR,
                value=value
            )
        )

    def output(self):
        """
        Main `Output`'s `output` method.
        First we get entries to add their outputs to the results and then
        calls specific `output` methods based (for instance) on preferred format.
        """
        if self._format_to_json:
            self._output_json()
        else:
            # Iterate through the entries and run their output method to add their content.
            for entry in self._entries:
                entry.output(self)
            self._output_text()

    def _output_json(self):
        """
        Finally outputs entries data to JSON format.
        See `archey.api.JSONAPI` for further documentation.
        """
        print(
            API(self._entries).json_serialization(
                indent=(self._format_to_json - 1)
            )
        )

    def _output_text(self):
        """
        Finally render the output entries.
        It handles text centering additionally to value and colors replacing.
        """
        # Let's copy the logo (so we don't modify the constant!)
        logo = LOGOS_DICT[self._distribution].copy()
        logo_width = get_logo_width(logo, len(self._colors_palette))

        # Let's center the entries and the logo (handles odd numbers)
        height_diff = len(logo) - len(self._results)
        if height_diff >= 0:
            self._results[0:0] = [''] * (height_diff // 2)
            self._results.extend([''] * (len(logo) - len(self._results)))
        else:
            colored_empty_line = [str(self._colors_palette[0]) + ' ' * logo_width]
            logo[0:0] = colored_empty_line * (-height_diff // 2)
            logo.extend(colored_empty_line * (len(self._results) - len(logo)))

        text_wrapper = TextWrapper(
            width=(get_terminal_size().columns - logo_width),
            expand_tabs=False,
            replace_whitespace=False,
            drop_whitespace=False,
            break_on_hyphens=False,
            max_lines=1,
            placeholder='...'
        )
        placeholder_length = len(text_wrapper.placeholder)

        # Using `TextWrapper`, shortens each entry to remove any line overlapping
        for i, entry in enumerate(self._results):
            # Shortens the entry according to the terminal width.
            # We have to remove any ANSI color, or the result would be skewed.
            wrapped_entry = text_wrapper.fill(Colors.remove_colors(entry))
            placeholder_offset = (
                placeholder_length if wrapped_entry.endswith(text_wrapper.placeholder) else 0
            )

            # By using previous positions, re-inserts ANSI colors back in the wrapped string.
            for color_match in ANSI_ECMA_REGEXP.finditer(entry):
                match_index = color_match.start()
                if match_index <= len(wrapped_entry) - placeholder_offset:
                    wrapped_entry = (wrapped_entry[:match_index]
                                     + color_match.group()
                                     + wrapped_entry[match_index:])

            # Add a color reset character before the placeholder (if any).
            # Rationale :
            # We cannot set `Colors.CLEAR` in the placeholder as it would skew its internals.
            if placeholder_offset:
                wrapped_entry = (wrapped_entry[:-placeholder_length]
                                 + str(Colors.CLEAR)
                                 + wrapped_entry[-placeholder_length:])

            self._results[i] = wrapped_entry

        # Merge entry results to the distribution logo.
        logo_with_entries = os.linesep.join([
            logo_part + entry_part
            for logo_part, entry_part in zip(logo, self._results)
        ])

        try:
            print(
                logo_with_entries.format(
                    c=self._colors_palette
                ) + str(Colors.CLEAR)
            )
        except UnicodeError:
            sys.exit(
                """\
Your locale or TTY does not seem to support UTF-8 encoding.
Please disable Unicode within your configuration file.\
""")



"""User session detection class"""

import os

from subprocess import CalledProcessError, check_output

from archey.entry import Entry


class User(Entry):
    """Retrieves the session name of the current logged in user"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = os.getenv('USER')
        if not self.value:
            try:
                self.value = check_output(
                    ['id', '-u', '-n'],
                    universal_newlines=True
                ).rstrip()
            except CalledProcessError:
                # Should not occur, but who knows ?
                pass



"""Test module for Archey's shell detection module"""

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, patch

from archey.entries.shell import Shell
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestShellEntry(unittest.TestCase):
    """
    For this entry, we'll just verify that the output is non-null.
    """
    @patch(
        'archey.entries.shell.os.getenv',
        return_value='SHELL'
    )
    def test_getenv(self, _):
        """Simple mock, simple test"""
        self.assertEqual(Shell().value, 'SHELL')

    @patch(
        'archey.entries.shell.os.getenv',
        return_value=None
    )
    @patch(
        'archey.entries.shell.os.getuid',  # We DO NOT HAVE TO mock this call.
        return_value=1000
    )
    @patch(
        'archey.entries.shell.check_output',
        return_value="USERNAME:x:1000:1000:User Name,,,:/home/user:/bin/bash\n"
    )
    def test_getent_call(self, _, __, ___):
        """Mock `getent` returned value and check the correct assignment"""
        self.assertEqual(Shell().value, '/bin/bash')

    @patch(
        'archey.entries.shell.os.getenv',
        return_value=None
    )
    @patch(
        'archey.entries.shell.check_output',
        side_effect=CalledProcessError(2, 'getent')
    )
    @HelperMethods.patch_clean_configuration
    def test_config_fall_back(self, _, __):
        """`id` fails, but Archey must not !"""
        shell = Shell()

        output_mock = MagicMock()
        shell.output(output_mock)

        self.assertIsNone(shell.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )

if __name__ == '__main__':
    unittest.main()



"""Elementary OS logo"""

ELEMENTARY = [
    """{c[0]}          eeeeeeeeeeeeeeeee          """,
    """{c[0]}       eeeeeeeeeeeeeeeeeeeeeee       """,
    """{c[0]}     eeeee  eeeeeeeeeeee   eeeee     """,
    """{c[0]}   eeee   eeeee       eee     eeee   """,
    """{c[0]}  eeee   eeee          eee     eeee  """,
    """{c[0]} eee    eee            eee       eee """,
    """{c[0]} eee   eee            eee        eee """,
    """{c[0]} ee    eee           eeee       eeee """,
    """{c[0]} ee    eee         eeeee      eeeeee """,
    """{c[0]} ee    eee       eeeee      eeeee ee """,
    """{c[0]} eee   eeee   eeeeee      eeeee  eee """,
    """{c[0]} eee    eeeeeeeeee     eeeeee    eee """,
    """{c[0]}  eeeeeeeeeeeeeeeeeeeeeeee    eeeee  """,
    """{c[0]}   eeeeeeee eeeeeeeeeeee      eeee   """,
    """{c[0]}     eeeee                 eeeee     """,
    """{c[0]}       eeeeeee         eeeeeee       """,
    """{c[0]}          eeeeeeeeeeeeeeeee          """
]



"""Archey configuration module"""

import os
import sys
import json
from copy import deepcopy

from archey.constants import DEFAULT_CONFIG
from archey.singleton import Singleton


class Configuration(metaclass=Singleton):
    """
    Values present in `archey.constants.DEFAULT_CONFIG` dictionary are required.
    New optional values may be added with `update_recursive` method.

    If a `config_path` is passed during instantiation, it will be loaded.
    """
    def __init__(self, config_path=None):
        # Deep-copy `DEFAULT_CONFIG` so we have a local copy to safely mutate.
        self._config = deepcopy(DEFAULT_CONFIG)

        # Let's "save" `STDERR` file descriptor for `suppress_warnings` option
        self._stderr = sys.stderr

        # If a `config_path` has been specified, (try to) load it directly.
        if config_path:
            self._load_configuration(config_path)
        # If not, load each (optional) configuration file in a "regular" order.
        else:
            self._load_configuration('/etc/archey4/')
            self._load_configuration(os.path.expanduser('~/.config/archey4/'))
            self._load_configuration(os.path.dirname(os.path.realpath(__file__)))

    def get(self, key, default=None):
        """
        A binding method to imitate the `dict.get()` behavior.
        """
        return self._config.get(key, default)

    def _load_configuration(self, path):
        """
        A method handling configuration loading from a JSON file.
        It will try to load any `config.json` present under `path`.
        """
        # If a previous configuration file has denied overriding...
        if not self.get('allow_overriding'):
            #  ... don't load this one.
            return

        # If the specified `path` is a directory, append the file name we are looking for.
        if os.path.isdir(path):
            path = os.path.join(path, 'config.json')

        try:
            with open(path) as f_config:
                self.update_recursive(self._config, json.load(f_config))
        except FileNotFoundError:
            return
        # For backward compatibility with Python versions prior to 3.5.0
        #   we use `ValueError` instead of `json.JSONDecodeError`.
        except ValueError as value_error:
            print('Warning: {0} ({1})'.format(value_error, path), file=sys.stderr)
            return

        # If the user does not want any warning to appear : 2> /dev/null
        if self.get('suppress_warnings'):
            # One more if statement to avoid multiple `open` calls.
            if sys.stderr == self._stderr:
                sys.stderr = open(os.devnull, 'w')
        else:
            self._close_and_restore_sys_stderr()

    @classmethod
    def update_recursive(cls, old_dict, new_dict):
        """
        A method for recursively merging dictionaries as `dict.update()` is not able to do this.
        Original snippet taken from here : <https://gist.github.com/angstwad/bf22d1822c38a92ec0a9>
        """
        for key, value in new_dict.items():
            if key in old_dict \
                and isinstance(old_dict[key], dict) \
                and isinstance(value, dict):
                cls.update_recursive(old_dict[key], value)
            else:
                old_dict[key] = value

    def _close_and_restore_sys_stderr(self):
        """If modified, close current and restore `sys.stderr` to its original file descriptor"""
        if sys.stderr != self._stderr:
            sys.stderr.close()
            sys.stderr = self._stderr

    def __del__(self):
        self._close_and_restore_sys_stderr()

    def __iter__(self):
        """When used as an iterator, directly yield `_config` elements"""
        return iter(self._config.items())



"""Public IP address detection class"""

from socket import timeout as SocketTimeoutError
from subprocess import check_output, DEVNULL, TimeoutExpired, CalledProcessError
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from archey.entry import Entry


class WanIp(Entry):
    """Uses different ways to retrieve the public IPv{4,6} addresses"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = []

        self._retrieve_ipv4_address()

        # IPv6 address retrieval (unless the user doesn't want it).
        if self._configuration.get('ip_settings')['wan_ip_v6_support']:
            self._retrieve_ipv6_address()

        self.value = list(filter(None, self.value))

    def _retrieve_ipv4_address(self):
        try:
            ipv4_addr = check_output(
                [
                    'dig', '+short', '-4', 'A', 'myip.opendns.com',
                    '@resolver1.opendns.com'
                ],
                timeout=self._configuration.get('timeout')['ipv4_detection'],
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except (FileNotFoundError, TimeoutExpired, CalledProcessError):
            try:
                ipv4_addr = urlopen(
                    'https://v4.ident.me/',
                    timeout=self._configuration.get('timeout')['ipv4_detection']
                )
            except (HTTPError, URLError, SocketTimeoutError):
                # The machine does not seem to be connected to Internet...
                return

            ipv4_addr = ipv4_addr.read().decode().strip()

        self.value.append(ipv4_addr)

    def _retrieve_ipv6_address(self):
        try:
            ipv6_addr = check_output(
                [
                    'dig', '+short', '-6', 'AAAA', 'myip.opendns.com',
                    '@resolver1.ipv6-sandbox.opendns.com'
                ],
                timeout=self._configuration.get('timeout')['ipv6_detection'],
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except (FileNotFoundError, TimeoutExpired, CalledProcessError):
            try:
                response = urlopen(
                    'https://v6.ident.me/',
                    timeout=self._configuration.get('timeout')['ipv6_detection']
                )
            except (HTTPError, URLError, SocketTimeoutError):
                # It looks like this machine doesn't have any IPv6 address...
                # ... or is not connected to Internet.
                return

            ipv6_addr = response.read().decode().strip()

        self.value.append(ipv6_addr)


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting our list of IP addresses."""
        # If we found IP addresses, join them together nicely.
        # If not, fall-back on the "No address" string.
        output.append(
            self.name,
            ', '.join(self.value) or self._configuration.get('default_strings')['no_address']
        )



"""Terminal detection class"""

import os
import re

from archey.colors import Colors, NO_COLOR
from archey.entry import Entry


# We detect a terminal by using the following three constants in the order below:
# First, we try using the value in the `TERM_PROGRAM` environment variable.
# Then, we use `COLORTERM_DICT` to try matching a value with the `COLORTERM` one.
# Third, we use `TERM_DICT` to try matching a value with the `TERM` one.
# Finally, we fall back to custom environment variables defined in `ENV_DICT`.
# If none of the above tests find a value, we use whichever value was defined in `$TERM`.

# All of the keys in `COLORTERM_DICT` and `TERM_DICT` are matched as regular expressions...
#   (using `re.match`, i.e. attempting to match once at the beginning of the string), so be careful!


# This dictionary contains values for the `COLORTERM` environment variable for terminal emulators
#   which do not propagate any other usable environment variable.
# If `COLORTERM` matches one of these values, a normalization is performed with its corresponding
#   value (i.e. the respective terminal emulator).
# If the variable does not match any keys in this dictionary, it is ignored.
COLORTERM_DICT = {
    r'kmscon': 'KMSCON',
    r'rxvt': 'rxvt',
}

# This dictionary contains values for the `TERM` environment variable for terminal emulators
#   which do not propagate any other usable environment variable.
# If `TERM` matches one of these values, a normalization is performed with its corresponding
#   value (i.e. the respective terminal emulator).
# If the variable does not match any keys in this dictionary, it is ignored,
#   UNLESS it does not begin with `xterm`, at which point its value is taken as the terminal in use.
#   This behavior can be overridden by specifying its exact match here.
TERM_DICT = {
    r'xterm-termite': 'Termite',
}

# This dictionary contains environment variables used to detect terminal emulators...
#   which do not propagate any usable `COLORTERM`, `TERM`, or `TERM_PROGRAM`.
# When a key is found in environment, a normalization is performed with its corresponding value.
ENV_DICT = {
    'ALACRITTY_LOG': 'Alacritty',
    'GNOME_TERMINAL_SCREEN': 'GNOME Terminal',
    'GUAKE_TAB_UUID': 'Guake',
    'KITTY_WINDOW_ID': 'Kitty',
    'KONSOLE_VERSION': 'Konsole',
    'MLTERM': 'MLTERM',
    'TERMINATOR_UUID': 'Terminator',
}


class Terminal(Entry):
    """
    Simple terminal detection based on the `TERM` environment variable.
    It also displays the colors palette afterwards.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = self._detect_terminal_emulator()

    def _get_colors_palette(self):
        """Build and return a 8-color palette, with Unicode characters if allowed"""
        # On systems with non-Unicode locales, we imitate '\u2588' character
        # ... with '#' to display the terminal colors palette.
        use_unicode = self._configuration.get('colors_palette')['use_unicode']

        return ' '.join([
            '{normal}{character}{bright}{character}{clear}'.format(
                normal=Colors((0, i)),
                bright=Colors((1, i)),
                character=('\u2588' if use_unicode else '#'),
                clear=Colors.CLEAR
            ) for i in range(37, 30, -1)
        ])

    @staticmethod
    def _detect_terminal_emulator():
        """Try to detect current terminal emulator based on various environment variables"""
        # At first, try to honor `TERM_PROGRAM` environment variable.
        # See <https://github.com/Maximus5/ConEmu/issues/1837#issuecomment-469199525>.
        env_term_program = os.getenv('TERM_PROGRAM')
        if env_term_program:
            return env_term_program

        # Second, check if we have any matches as defined in our `COLORTERM` constant dict.
        env_colorterm = os.getenv('COLORTERM')
        if env_colorterm:
            for env_value_re, normalized_name in COLORTERM_DICT.items():
                if re.match(env_value_re, env_colorterm):
                    return normalized_name

        # Third, check if we have any matches defined in our `TERM` constant dict.
        env_term = os.getenv('TERM')
        if env_term:
            for env_value_re, normalized_name in TERM_DICT.items():
                if re.match(env_value_re, env_term):
                    return normalized_name

            # If we didn't find any match and `TERM` is set to "something special", honor it.
            if not env_term.startswith('xterm'):
                return env_term

        # If not, try to find a "known identifier" and perform name normalization...
        for env_var, normalized_name in ENV_DICT.items():
            if env_var in os.environ:
                return normalized_name

        # When nothing of the above matched, falls-back on the regular `TERM` environment variable.
        # Note : It _might_ be `None` in very specific environments.
        return env_term


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting with colors palette"""
        text_output = (self.value or self._configuration.get('default_strings')['not_detected'])
        if not NO_COLOR:
            text_output += ' ' + self._get_colors_palette()

        output.append(self.name, text_output)



"""Distribution and architecture detection class"""

from subprocess import check_output

from archey.distributions import Distributions
from archey.entry import Entry


class Distro(Entry):
    """Uses `distro` module and `uname` system program to format `${DISTRO} [${ARCH}]` string"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        distro_name = Distributions.get_distro_name()
        if not distro_name:
            distro_name = self._fetch_android_release()

        self.value = {
            'name': distro_name,
            'arch': self._fetch_architecture()
        }

    @staticmethod
    def _fetch_architecture():
        """Simple wrapper to `uname -m` returning the current system architecture"""
        return check_output(
            ['uname', '-m'],
            universal_newlines=True
        ).rstrip()

    @staticmethod
    def _fetch_android_release():
        """Simple method to fetch current release on Android systems"""
        try:
            release = check_output(
                ['getprop', 'ro.build.version.release'],
                universal_newlines=True
            ).rstrip()
        except FileNotFoundError:
            return None

        return 'Android {0}'.format(release)


    def output(self, output):
        output.append(
            self.name,
            '{0} [{1}]'.format(
                (self.value['name'] or self._configuration.get('default_strings')['not_detected']),
                self.value['arch']
            )
        )



"""Test module for Archey's installed system packages detection module"""

import unittest
from unittest.mock import DEFAULT as DEFAULT_SENTINEL, MagicMock, patch

from archey.constants import DEFAULT_CONFIG
from archey.test.entries import HelperMethods
from archey.entries.packages import Packages


class TestPackagesEntry(unittest.TestCase):
    """
    Here, we mock the `check_output` calls and check afterwards
      that the outputs are correct.
    Sorry for the code style, mocking this class is pretty boring.

    Note: Due to the presence of trailing spaces, we may have to manually
            "inject" the OS line separator in some mocked outputs.
    """
    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
sqlite-libs-3.30.1-r1 x86_64 {{sqlite}} (Public-Domain) [installed]
musl-1.1.24-r2 x86_64 {{musl}} (MIT) [installed]
libbz2-1.0.8-r1 x86_64 {{bzip2}} (bzip2-1.0.6) [installed]
gdbm-1.13-r1 x86_64 {{gdbm}} (GPL) [installed]
ncurses-libs-6.1_p20200118-r3 x86_64 {{ncurses}} (MIT) [installed]
zlib-1.2.11-r3 x86_64 {{zlib}} (Zlib) [installed]
apk-tools-2.10.4-r3 x86_64 {{apk}-tools} (GPL2) [installed]
readline-8.0.1-r0 x86_64 {{readline}} (GPL-2.0-or-later) [installed]
""")
    def test_match_with_apk(self, check_output_mock):
        """Simple test for the APK packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('apk')

        self.assertEqual(Packages().value, 8)

# `apt` is disabled for now.
#    @patch(
#        'archey.entries.packages.check_output',
#        return_value="""\
#accountsservice/stable,now 0.6.45-2 amd64 [installed,automatic]
#acl/stable,now 2.2.53-4 amd64 [installed,automatic]
#adb/stable,now 1:8.1.0+r23-5 amd64 [installed]
#adduser/stable,now 3.118 all [installed]
#adwaita-icon-theme/stable,now 3.30.1-1 all [installed,automatic]
#albatross-gtk-theme/stable,now 1.7.4-1 all [installed,automatic]
#alsa-utils/stable,now 1.1.8-2 amd64 [installed,automatic]
#""")
#    def test_match_with_apt(self, check_output_mock):
#        """Simple test for the APT packages manager"""
#        check_output_mock.side_effect = self._check_output_side_effect('apt')

#        self.assertEqual(Packages().value, 7)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
Installed Packages
GConf2.x86_64                  3.2.6-17.fc26           @@commandline
GeoIP.x86_64                   1.6.11-1.fc26           @@commandline
GeoIP-GeoLite-data.noarch      2017.07-1.fc26          @@commandline
GraphicsMagick.x86_64          1.3.26-3.fc26           @@commandline
""")
    def test_match_with_dnf(self, check_output_mock):
        """Simple test for the DNF packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('dnf')

        self.assertEqual(Packages().value, 4)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
accountsservice         install
acl                     install
adb                     install
adduser                 install
adwaita-icon-theme      install
albatross-gtk-theme     deinstall
alien                   install
""")
    def test_match_with_dpkg(self, check_output_mock):
        """Simple test for the DPKG packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('dpkg')

        self.assertEqual(Packages().value, 6)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\

These are the packages that would be merged, in order:

Calculating dependencies  ... done!
[ebuild     U  ] sys-libs/glibc-2.25-r10 [2.25-r9]
[ebuild   R    ] sys-apps/busybox-1.28.0 \n\
[ebuild  N     ] sys-libs/libcap-2.24-r2  \
USE="pam -static-libs" ABI_X86="(64) -32 (-x32)" \n\
[ebuild     U  ] app-misc/pax-utils-1.2.2-r2 [1.1.7]
[ebuild   R    ] x11-misc/shared-mime-info-1.9 \n\

""")
    def test_match_with_emerge(self, check_output_mock):
        """Simple test for the Emerge packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('emerge')

        self.assertEqual(Packages().value, 5)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
nix-2.3.4
nss-cacert-3.49.2
python3-3.8.2
python3.8-pip-20.1
""")
    def test_match_with_nix_env(self, check_output_mock):
        """Simple test for the Emerge packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('nix-env')

        self.assertEqual(Packages().value, 4)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
acl 2.2.52-4
archey4 v4.3.3-1
archlinux-keyring 20180108-1
argon2 20171227-3
""")
    def test_match_with_pacman(self, check_output_mock):
        """Simple test for the Pacman packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('pacman')

        self.assertEqual(Packages().value, 4)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
bzip2-1.0.8         block-sorting file compressor, unencumbered
gettext-runtime-0.20.1p0 GNU gettext runtime libraries and programs
intel-firmware-20191115v0 microcode update binaries for Intel CPUs
libffi-3.2.1p5      Foreign Function Interface
libiconv-1.16p0     character set conversion library
python-3.7.4        interpreted object-oriented programming language
quirks-3.187        exceptions to pkg_add rules
sqlite3-3.29.0      embedded SQL implementation
xz-5.2.4            LZMA compression and decompression tools
"""
    )
    def test_match_with_pkg_info(self, check_output_mock):
        """Simple test for the OpenBSD `pkg_*` package manager"""
        check_output_mock.side_effect = self._check_output_side_effect('pkg_info')

        self.assertEqual(Packages().value, 9)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
gettext-runtime-0.20.1         GNU gettext runtime libraries and programs
indexinfo-0.3.1                Utility to regenerate the GNU info page index
libffi-3.2.1_3                 Foreign Function Interface
pkg-1.13.2                     Package manager
python-3.7_3,2                 "meta-port" for the default version of Python interpreter
python3-3_3                    The "meta-port" for version 3 of the Python interpreter
python37-3.7.7                 Interpreted object-oriented programming language
readline-8.0.4                 Library for editing command lines as they are typed
"""
    )
    def test_match_with_pkg(self, check_output_mock):
        """Simple test for the FreeBSD `pkg` package manager"""
        check_output_mock.side_effect = self._check_output_side_effect('pkg')

        self.assertEqual(Packages().value, 8)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
cdrecord-2.01-10.7.el5
bluez-libs-3.7-1.1
setarch-2.0-1.1
MySQL-client-3.23.57-1
""")
    def test_match_with_rpm(self, check_output_mock):
        """Simple test for the RPM packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('rpm')

        self.assertEqual(Packages().value, 4)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
Loaded plugins: fastestmirror, langpacks
Installed Packages
GConf2.x86_64                   3.2.6-8.el7         @base/$releasever
GeoIP.x86_64                    1.5.0-11.el7        @base            \n\
ModemManager.x86_64             1.6.0-2.el7         @base            \n\
ModemManager-glib.x86_64        1.6.0-2.el7         @base            \n\
""")
    def test_match_with_yum(self, check_output_mock):
        """Simple test for the Yum packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('yum')

        self.assertEqual(Packages().value, 4)

    @patch(
        'archey.entries.packages.check_output',
        return_value="""\
Loading repository data...
Reading installed packages...

S  | Name          | Summary                             | Type       \n\
---+---------------+-------------------------------------+------------
i+ | 5201          | Recommended update for xdg-utils    | patch      \n\
i  | GeoIP-data    | Free GeoLite country-data for GeoIP | package    \n\
i  | make          | GNU make                            | package    \n\
i  | GNOME Nibbles | Guide a worm around a maze          | application
i  | at            | A Job Manager                       | package    \n\
""")
    def test_match_with_zypper(self, check_output_mock):
        """Simple test for the Zypper packages manager"""
        check_output_mock.side_effect = self._check_output_side_effect('zypper')

        self.assertEqual(Packages().value, 5)

    @patch(
        'archey.entries.packages.PACKAGES_TOOLS',
        new=(
            {'cmd': ('pkg_tool_1')},
            {'cmd': ('pkg_tool_2'), 'skew': 2}
        )
    )
    @patch(
        'archey.entries.packages.check_output',
        side_effect=[
            """\
sample_package_1_1
sample_package_1_2
""",
            """\
  Incredible list of installed packages:
sample_package_2_1
sample_package_2_2

"""
        ]
    )
    def test_multiple_package_managers(self, _):
        """Simple test for multiple packages managers"""
        self.assertEqual(Packages().value, 4)

    @patch('archey.entries.packages.check_output')
    @HelperMethods.patch_clean_configuration
    def test_no_packages_manager(self, check_output_mock):
        """No packages manager is available at the moment..."""
        check_output_mock.side_effect = self._check_output_side_effect()

        packages = Packages()

        output_mock = MagicMock()
        packages.output(output_mock)

        self.assertIsNone(packages.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


    @staticmethod
    def _check_output_side_effect(pkg_manager_cmd=None):
        """Internal helper method to facilitate `check_output` mocking"""
        def _check_output(args, stderr, env, universal_newlines):  # pylint: disable=unused-argument
            """
            This closure is a drop-replacement for our `check_output` call.
            If the _called_ program is `pkg_manager_cmd`, patched `return_value` will be returned.
            If not, `FileNotFoundError` would be raised.
            """
            if args[0] == pkg_manager_cmd:
                return DEFAULT_SENTINEL

            raise FileNotFoundError

        return _check_output


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's uptime detection module"""

import unittest
from unittest.mock import mock_open, patch, MagicMock
from datetime import timedelta
from itertools import product

from archey.entries.uptime import Uptime


class TestUptimeEntry(unittest.TestCase):
    """
    Here, we mock the `open` call and imitate `/proc/uptime` content.
    """
    @patch(
        'archey.entries.uptime.open',
        mock_open(
            read_data='0.00 XXXX.XX\n'
        ),
        create=True
    )
    def test_warming_up(self):
        """Test when the device has just been started..."""
        uptime = Uptime()

        output_mock = MagicMock()
        uptime.output(output_mock)

        self.assertDictEqual(
            uptime.value,
            {
                'days': 0,
                'hours': 0,
                'minutes': 0,
                'seconds': 0
            }
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '< 1 minute'
        )

    @patch(
        'archey.entries.uptime.open',
        mock_open(
            read_data='120.25 XXXX.XX\n'
        ),
        create=True
    )
    def test_minutes_only(self):
        """Test when only minutes should be displayed"""
        output_mock = MagicMock()
        Uptime().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '2 minutes'
        )

    @patch(
        'archey.entries.uptime.open',
        mock_open(
            read_data='7260.50 XXXX.XX\n'
        ),
        create=True
    )
    def test_hours_and_minute(self):
        """Test when only hours AND minutes should be displayed"""
        output_mock = MagicMock()
        Uptime().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '2 hours and 1 minute'
        )

    @patch(
        'archey.entries.uptime.open',
        mock_open(
            read_data='90120.75 XXXX.XX\n'
        ),
        create=True
    )
    def test_day_and_hour_and_minutes(self):
        """Test when only days, hours AND minutes should be displayed"""
        output_mock = MagicMock()
        Uptime().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '1 day, 1 hour and 2 minutes'
        )

    @patch(
        'archey.entries.uptime.open',
        mock_open(
            read_data='259381.99 XXXX.XX\n'
        ),
        create=True
    )
    def test_days_and_minutes(self):
        """Test when only days AND minutes should be displayed"""
        uptime = Uptime()

        output_mock = MagicMock()
        uptime.output(output_mock)

        self.assertDictEqual(
            uptime.value,
            {
                'days': 3,
                'hours': 0,
                'minutes': 3,
                'seconds': 1
            }
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '3 days and 3 minutes'
        )

    @patch(
        'archey.entries.uptime.open',
        side_effect=PermissionError(),
        create=True
    )
    @patch(
        'archey.entries.uptime.time.CLOCK_BOOTTIME',  # Ensure first `getattr` call succeeds anyhow.
        create=True
    )
    @patch(
        'archey.entries.uptime.time.clock_gettime',
        return_value=1000,
        create=True
    )
    def test_clock_fallback(self, _, __, ___):
        """
        Test when we can't access /proc/uptime on Linux/macOS/BSD.
        We only test one clock as all clocks rely on the same built-in `time.clock_gettime` method.
        """
        output_mock = MagicMock()
        Uptime().output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '16 minutes'
        )

    @patch('archey.entries.uptime.check_output')
    def test_uptime_fallback(self, check_output_mock):
        """Test `uptime` command parsing"""
        # Create an uptime instance to perform testing.
        # It doesn't matter that its `__init__` will be called.
        uptime_inst = Uptime()

        # Keys: `uptime` outputs; values: expected `timedelta` instances.
        # We will test these with various time formats (and various numbers of users).
        # pylint: disable=line-too-long
        test_uptime_cases = {
            # Recently booted:
            '{time} up 0 sec, {user_loadavg}': timedelta(seconds=0),  # BSD, just booted
            '{time} up 1 sec, {user_loadavg}': timedelta(seconds=1),  # BSD, < 1 min uptime
            '{time} up 12 secs, {user_loadavg}': timedelta(seconds=12),  # BSD, < 1 min uptime
            '{time} up 0 min, {user_loadavg}': timedelta(minutes=0),  # Linux, < 1 min uptime
            '{time} up 1 min, {user_loadavg}': timedelta(minutes=1),
            # 1 min to 1 day
            '{time} up 12 mins, {user_loadavg}': timedelta(minutes=12),
            '{time} up 12 min, {user_loadavg}': timedelta(minutes=12),  # Variation without plural minutes
            '{time} up 1:00, {user_loadavg}': timedelta(hours=1),
            '{time} up 1:01, {user_loadavg}': timedelta(hours=1, minutes=1),
            '{time} up 1:23, {user_loadavg}': timedelta(hours=1, minutes=23),
            '{time} up 12:34, {user_loadavg}': timedelta(hours=12, minutes=34),
            # 1 day to 2 days
            '{time} up 1 day, 0 sec, {user_loadavg}': timedelta(days=1),  # BSD
            '{time} up 1 day, 1 sec, {user_loadavg}': timedelta(days=1, seconds=1),  # BSD
            '{time} up 1 day, 12 secs, {user_loadavg}': timedelta(days=1, seconds=12),  # BSD
            '{time} up 1 day, 0 min, {user_loadavg}': timedelta(days=1),  # Linux
            '{time} up 1 day, 1 min, {user_loadavg}': timedelta(days=1, minutes=1),
            '{time} up 1 day, 12 mins, {user_loadavg}': timedelta(days=1, minutes=12),
            '{time} up 1 day, 12 min, {user_loadavg}': timedelta(days=1, minutes=12),  # Variation without plural minutes
            '{time} up 1 day, 1:00, {user_loadavg}': timedelta(days=1, hours=1),
            '{time} up 1 day, 1:01, {user_loadavg}': timedelta(days=1, hours=1, minutes=1),
            '{time} up 1 day, 1:23, {user_loadavg}': timedelta(days=1, hours=1, minutes=23),
            '{time} up 1 day, 12:34, {user_loadavg}': timedelta(days=1, hours=12, minutes=34),
            # 2 days onwards
            '{time} up 12 days, 0 sec, {user_loadavg}': timedelta(days=12),  # BSD
            '{time} up 12 days, 1 sec, {user_loadavg}': timedelta(days=12, seconds=1),  # BSD
            '{time} up 12 days, 12 secs, {user_loadavg}': timedelta(days=12, seconds=12),  # BSD
            '{time} up 12 days, 0 min, {user_loadavg}': timedelta(days=12),  # Linux
            '{time} up 12 days, 1 min, {user_loadavg}': timedelta(days=12, minutes=1),
            '{time} up 12 days, 12 mins, {user_loadavg}': timedelta(days=12, minutes=12),
            '{time} up 12 day, 12 min, {user_loadavg}': timedelta(days=12, minutes=12),  # Variation without plural minutes
            '{time} up 12 days, 1:00, {user_loadavg}': timedelta(days=12, hours=1),
            '{time} up 12 days, 1:01, {user_loadavg}': timedelta(days=12, hours=1, minutes=1),
            '{time} up 12 days, 1:23, {user_loadavg}': timedelta(days=12, hours=1, minutes=23),
            '{time} up 12 days, 12:34, {user_loadavg}': timedelta(days=12, hours=12, minutes=34),
            # Very long uptimes - sanity check :)
            '{time} up 500 days, 0 sec, {user_loadavg}': timedelta(days=500),  # BSD
            '{time} up 500 days, 1 sec, {user_loadavg}': timedelta(days=500, seconds=1),  # BSD
            '{time} up 500 days, 12 secs, {user_loadavg}': timedelta(days=500, seconds=12),  # BSD
            '{time} up 500 days, 0 min, {user_loadavg}': timedelta(days=500),  # Linux
            '{time} up 500 days, 1 min, {user_loadavg}': timedelta(days=500, minutes=1),
            '{time} up 500 days, 12 mins, {user_loadavg}': timedelta(days=500, minutes=12),
            '{time} up 500 day, 12 min, {user_loadavg}': timedelta(days=500, minutes=12),  # Variation without plural minutes
            '{time} up 500 days, 1:00, {user_loadavg}': timedelta(days=500, hours=1),
            '{time} up 500 days, 1:01, {user_loadavg}': timedelta(days=500, hours=1, minutes=1),
            '{time} up 500 days, 1:23, {user_loadavg}': timedelta(days=500, hours=1, minutes=23),
            '{time} up 500 days, 12:34, {user_loadavg}': timedelta(days=500, hours=12, minutes=34),
        }
        # pylint: enable=line-too-long

        # Variations of the time in the `{time}` section.
        # These _should_ be avoided when we set the locale in `check_output`,
        # however let's check we can handle them anyway, just in case.
        time_variations = (
            '0:00', '9:43 ', '11:37  ', '19:21', '23:59',
            '12:00am', '1:10am', '1:10pm ', '6:43pm ', '8:26am ',
            '03:14:15', '09:26:12 ', '23:19:20  ',
            'nonsense_time  ', 'hopefully_works_anyway ',
            '  even with strange spacing!'
        )

        # Variations of the user count and load average section.
        # For this, we'll just combine user variations with a few load average variations.
        user_variations = (
            '1 user ', '1 user  ', ' 1 user, ', ' 1 user,  ',
            '2 users ', '2 users  ', '  2 users, ', '  2 users,  ',
            '15 users ', '15 users  ', ' 15 users, ', ' 15 users,  ',
            '150 users ', '150 users  ', '150 users, ', '150 users,  ',
        )
        loadavg_variations = (
            'load averages: 1.95 1.28 2.10',
            'load average: 0.13, 0.17, 0.13',
            'we never match this part so the content here',
            '  should not affect our parsing'
        )
        user_loadavg_variations = [
            user + loadavg
            for user in user_variations
            for loadavg in loadavg_variations
        ]

        for uptime_output, expected_delta in test_uptime_cases.items():
            # We use `itertools.product` to get the permutations of our variations
            # since there are a lot of them! (a list comprehension would be slower)
            for variations in product(time_variations, user_loadavg_variations):
                check_output_mock.return_value = uptime_output.format(
                    time=variations[0],
                    user_loadavg=variations[1]
                )
                self.assertEqual(
                    uptime_inst._parse_uptime_cmd(),  # pylint: disable=protected-access
                    expected_delta,
                    msg='`uptime` output: "{0}"'.format(
                        uptime_output.format(
                            time=variations[0],
                            user_loadavg=variations[1]
                        )
                    )
                )

    @patch(
        'archey.entries.uptime.open',
        side_effect=PermissionError(),
        create=True
    )
    @patch(
        'archey.entries.uptime.check_output',
        side_effect=FileNotFoundError()
    )
    @patch.object(
        Uptime,
        '_clock_uptime',
        side_effect=RuntimeError()
    )
    def test_procps_missing(self, _, __, ___):
        """Test `uptime` failure (program exit) when no uptime sources are available"""
        self.assertRaises(SystemExit, Uptime)


if __name__ == '__main__':
    unittest.main()
